import csv
import sys
import threading
import traceback
from collections import defaultdict
from datetime import datetime as Datetime
from decimal import Decimal
from itertools import combinations
from numbers import Number

from dateutil.relativedelta import relativedelta

from flask import g, redirect, request

from sqlalchemy import or_, select
from sqlalchemy.orm import joinedload, subqueryload
from sqlalchemy.sql.expression import null

from werkzeug.exceptions import BadRequest

from zish import ZishException


from chellow.dloads import open_file
from chellow.e.computer import SupplySource, contract_func
from chellow.models import (
    Batch,
    Bill,
    Contract,
    Element,
    Era,
    Llfc,
    MtcParticipant,
    RSession,
    RegisterRead,
    ReportRun,
    Supply,
    User,
)
from chellow.utils import (
    HH,
    csv_make_val,
    date_format,
    hh_max,
    hh_min,
    hh_range,
    parse_mpan_core,
    req_date,
    req_int,
    req_str,
    to_utc,
)


def _add_gap_hh(gaps, hh_start, gap_type):
    try:
        hh = gaps[hh_start]
        match (hh, gap_type):
            case ("middle", _):
                pass
            case (_, "middle"):
                gaps[hh_start] = "middle"
            case ("start", "start"):
                pass
            case ("start", "finish"):
                gaps[hh_start] = "start_finish"
            case ("finish", "finish"):
                pass
            case ("finish", "start"):
                gaps[hh_start] = "start_finish"
            case ("start_finish", "finish"):
                pass
            case ("start_finish", "start"):
                pass
            case _:
                raise BadRequest(f"Gap combination ({hh}, {gap_type}) not recognized.")

    except KeyError:
        hh = gaps[hh_start] = gap_type


def _add_gap(caches, gaps, start_date, finish_date):
    hhs = hh_range(caches, start_date, finish_date)
    _add_gap_hh(gaps, hhs[0], "start")
    _add_gap_hh(gaps, hhs[-1] + HH, "finish")
    for hh_start in hhs[1:]:
        _add_gap_hh(gaps, hh_start, "middle")


def find_gaps(gaps):
    if len(gaps) > 0:
        gap_start = None
        for ghh, gtype in sorted(gaps.items()):
            if "finish" in gtype:
                yield gap_start, ghh - HH
            if "start" in gtype:
                gap_start = ghh


def content(
    batch_id,
    bill_id,
    contract_id,
    start_date,
    finish_date,
    user_id,
    mpan_cores,
    fname_additional,
    report_run_id,
):
    caches = {}
    tmp_file = sess = supply_id = None
    forecast_date = to_utc(Datetime.max)

    try:
        with RSession() as sess:
            user = User.get_by_id(sess, user_id)
            tmp_file = open_file(
                f"bill_check_{fname_additional}.csv", user, mode="w", newline=""
            )
            writer = csv.writer(tmp_file, lineterminator="\n")

            bills_q = (
                select(Bill)
                .order_by(Bill.supply_id, Bill.reference)
                .options(
                    joinedload(Bill.supply),
                    subqueryload(Bill.reads).joinedload(RegisterRead.present_type),
                    subqueryload(Bill.reads).joinedload(RegisterRead.previous_type),
                    joinedload(Bill.batch),
                )
            )

            if len(mpan_cores) > 0:
                mpan_cores = list(map(parse_mpan_core, mpan_cores))
                supply_ids = sess.scalars(
                    select(Era.supply_id)
                    .where(
                        or_(
                            Era.imp_mpan_core.in_(mpan_cores),
                            Era.exp_mpan_core.in_(mpan_cores),
                        )
                    )
                    .distinct()
                ).all()

                bills_q = bills_q.join(Supply).where(Supply.id.in_(supply_ids))

            if batch_id is not None:
                batch = Batch.get_by_id(sess, batch_id)
                bills_q = bills_q.where(Bill.batch == batch)
                contract = batch.contract
            elif bill_id is not None:
                bill = Bill.get_by_id(sess, bill_id)
                bills_q = bills_q.where(Bill.id == bill.id)
                contract = bill.batch.contract
            elif contract_id is not None:
                contract = Contract.get_by_id(sess, contract_id)
                bills_q = bills_q.join(Batch).where(
                    Batch.contract == contract,
                    Bill.start_date <= finish_date,
                    Bill.finish_date >= start_date,
                )

            vbf = contract_func(caches, contract, "virtual_bill")
            if vbf is None:
                raise BadRequest(
                    f"The contract {contract.name} doesn't have a function "
                    f"virtual_bill."
                )

            virtual_bill_titles_func = contract_func(
                caches, contract, "virtual_bill_titles"
            )
            if virtual_bill_titles_func is None:
                raise BadRequest(
                    f"The contract {contract.name} doesn't have a function "
                    f"virtual_bill_titles."
                )
            virtual_bill_titles = virtual_bill_titles_func()

            titles = []
            header_titles = [
                "imp_mpan_core",
                "exp_mpan_core",
                "site_code",
                "site_name",
                "period_start",
                "period_finish",
                "actual_net_gbp",
                "virtual_net_gbp",
                "difference_net_gbp",
            ]
            titles.extend(header_titles)
            for t in virtual_bill_titles:
                if t not in ("net-gbp", "vat-gbp", "gross-gbp"):
                    titles.append("actual-" + t)
                    titles.append("virtual-" + t)
                    if t.endswith("-gbp"):
                        titles.append("difference-" + t)

            writer.writerow(titles)

            bill_map = defaultdict(set, {})
            for bill in sess.scalars(bills_q):
                bill_map[bill.supply.id].add(bill.id)

            for supply_id, bill_ids in bill_map.items():
                for data in _process_supply(
                    sess, caches, supply_id, bill_ids, forecast_date, contract, vbf
                ):
                    vals = {}
                    for title in header_titles:
                        vals[title] = data[title]
                    for el_name, el in data["elements"].items():
                        for part_name, part in el["parts"].items():
                            for typ, value in part.items():
                                vals[f"{typ}-{el_name}-{part_name}"] = value

                    writer.writerow(csv_make_val(vals.get(title)) for title in titles)
                    ReportRun.w_insert_row(
                        report_run_id,
                        "",
                        titles,
                        vals,
                        {"is_checked": False},
                        data=data,
                    )
        ReportRun.w_update(report_run_id, "finished")

    except BadRequest as e:
        if supply_id is None:
            prefix = "Problem: "
        else:
            prefix = f"Problem with supply {supply_id}:"
        tmp_file.write(prefix + e.description)
        ReportRun.w_update(report_run_id, "problem")
    except BaseException:
        msg = traceback.format_exc()
        sys.stderr.write(msg + "\n")
        if supply_id is None:
            prefix = "Problem: "
        else:
            prefix = f"Problem with supply {supply_id}:"
        tmp_file.write(prefix + msg)
        ReportRun.w_update(report_run_id, "interrupted")
    finally:
        if tmp_file is not None:
            tmp_file.close()


def do_get(sess):
    return do_post(sess)


def do_post(sess):
    batch_id = bill_id = contract_id = start_date = finish_date = None
    if "mpan_cores" in request.values:
        mpan_cores = req_str("mpan_cores").splitlines()
    else:
        mpan_cores = []

    fname_additional = ""

    if "batch_id" in request.values:
        batch_id = req_int("batch_id")
        batch = Batch.get_by_id(sess, batch_id)
        fname_additional = f"_batch_{batch.reference}"
    elif "bill_id" in request.values:
        bill_id = req_int("bill_id")
        bill = Bill.get_by_id(sess, bill_id)
        fname_additional = "bill_" + str(bill.id)
    elif "contract_id" in request.values:
        contract_id = req_int("contract_id")
        contract = Contract.get_by_id(sess, contract_id)

        start_date = req_date("start_date")
        finish_date = req_date("finish_date")

        s = ["contract", str(contract.id)]
        for dt in (start_date, finish_date):
            s.append(date_format(dt).replace(" ", "T").replace(":", ""))
        fname_additional = "_".join(s)
    else:
        raise BadRequest(
            "The bill check needs a batch_id, a bill_id or a start_date and "
            "finish_date."
        )

    report_run = ReportRun.insert(
        sess,
        "bill_check",
        g.user,
        fname_additional,
        {},
    )
    sess.commit()

    args = (
        batch_id,
        bill_id,
        contract_id,
        start_date,
        finish_date,
        g.user.id,
        mpan_cores,
        fname_additional,
        report_run.id,
    )
    threading.Thread(target=content, args=args).start()
    return redirect(f"/report_runs/{report_run.id}", 303)


def _get_bill_status(sess, bill_statuses, bill):
    try:
        bill_status = bill_statuses[bill.id]
    except KeyError:
        covered_bills = dict(
            (b.id, b)
            for b in sess.scalars(
                select(Bill)
                .join(Batch)
                .join(Contract)
                .where(
                    Bill.supply == bill.supply,
                    Bill.start_date <= bill.finish_date,
                    Bill.finish_date >= bill.start_date,
                    Contract.market_role == bill.batch.contract.market_role,
                )
                .order_by(Bill.start_date, Bill.issue_date)
            )
        )
        while True:
            to_del = None
            for a, b in combinations(covered_bills.values(), 2):
                if all(
                    (
                        a.start_date == b.start_date,
                        a.finish_date == b.finish_date,
                        a.net == -1 * b.net,
                        a.vat == -1 * b.vat,
                        a.gross == -1 * b.gross,
                    )
                ):
                    to_del = (a.id, b.id)
                    break
            if to_del is None:
                break
            else:
                for k in to_del:
                    del covered_bills[k]
                    bill_statuses[k] = None

        for k, v in covered_bills.items():
            bill_statuses[k] = v

        bill_status = bill_statuses[bill.id]

    return bill_status


def _process_period(
    sess,
    caches,
    supply,
    contract,
    bill_statuses,
    forecast_date,
    vbf,
    period_start,
    period_finish,
):
    actual_elems = {}
    vels = {}
    val_elems = {}
    virtual_bill = {"problem": "", "elements": vels}
    market_role_code = contract.market_role.code

    vals = {
        "supply_id": supply.id,
        "period_start": period_start,
        "period_finish": period_finish,
        "contract_id": contract.id,
        "contract_name": contract.name,
        "market_role_code": contract.market_role.code,
        "elements": val_elems,
        "virtual_net_gbp": 0,
        "actual_net_gbp": 0,
        "actual_bills": [],
        "problem": "",
    }

    for bill in sess.scalars(
        select(Bill)
        .join(Batch)
        .where(
            Bill.supply == supply,
            Bill.start_date <= period_finish,
            Bill.finish_date >= period_start,
            Batch.contract == contract,
        )
    ):
        if _get_bill_status(sess, bill_statuses, bill) is not None:
            actual_bill = {
                "id": bill.id,
                "start_date": bill.start_date,
                "finish_date": bill.finish_date,
                "problem": "",
                "net": bill.net,
                "vat": bill.vat,
                "gross": bill.gross,
                "kwh": bill.kwh,
                "breakdown": bill.breakdown,
                "batch_id": bill.batch_id,
                "batch_reference": bill.batch.reference,
            }

            read_dict = {}
            for read in bill.reads:
                gen_start = read.present_date.replace(hour=0).replace(minute=0)
                gen_finish = gen_start + relativedelta(days=1) - HH
                msn_match = False
                read_msn = read.msn
                for read_era in supply.find_eras(sess, gen_start, gen_finish):
                    if read_msn == read_era.msn:
                        msn_match = True
                        break

                if not msn_match:
                    virtual_bill["problem"] += (
                        f"The MSN {read_msn} of the register read {read.id} "
                        f"doesn't match the MSN of the era."
                    )

                for dt, typ in [
                    (read.present_date, read.present_type),
                    (read.previous_date, read.previous_type),
                ]:
                    key = f"{dt}-{read.msn}"
                    try:
                        if typ != read_dict[key]:
                            virtual_bill[
                                "problem"
                            ] += f" Reads taken on {dt} have differing read types."
                    except KeyError:
                        read_dict[key] = typ

            element_net = sum(el.net for el in bill.elements)
            vals["actual_bills"].append(actual_bill)
            if element_net != bill.net:
                actual_bill["problem"] += (
                    f"The Net GBP total of the elements is {element_net} doesn't "
                    f"match the bill Net GBP value of {bill.net}. "
                )
            if bill.gross != bill.vat + bill.net:
                actual_bill["problem"] += (
                    f"The Gross GBP ({bill.gross}) of the bill isn't equal to "
                    f"the Net GBP ({bill.net}) + VAT GBP ({bill.vat}) of the bill."
                )

            vat_net = Decimal("0.00")
            vat_vat = Decimal("0.00")

            try:
                bd = bill.bd

                if "vat" in bd:
                    for vat_percentage, vat_vals in bd["vat"].items():
                        vat_net += vat_vals["net"]
                        vat_vat += vat_vals["vat"]
            except ZishException as e:
                actual_bill["problem"] += f"Problem parsing the breakdown: {e}"

            if vat_net != bill.net:
                actual_bill["problem"] += (
                    f"The total 'net' {vat_net} in the VAT breakdown doesn't "
                    f"match the 'net' {bill.net} of the bill."
                )
            if vat_vat != bill.vat:
                actual_bill["problem"] += (
                    f"The total VAT {vat_vat} in the VAT breakdown doesn't "
                    f"match the VAT {bill.vat} of the bill."
                )

            if len(actual_bill["problem"]) > 0:
                vals["problem"] += "Bills have problems. "

    for element in sess.scalars(
        select(Element)
        .join(Bill)
        .join(Batch)
        .where(
            Bill.supply == supply,
            Element.start_date <= period_finish,
            Element.finish_date >= period_start,
            Batch.contract == contract,
        )
    ):
        if _get_bill_status(sess, bill_statuses, element.bill) is None:
            continue

        try:
            actual_elem = actual_elems[element.name]
        except KeyError:
            actual_elem = actual_elems[element.name] = {
                "parts": {"gbp": Decimal("0.00")},
                "elements": [],
            }
        parts = actual_elem["parts"]
        actual_elem["elements"].append(
            {
                "id": element.id,
                "start_date": element.start_date,
                "finish_date": element.finish_date,
                "net": element.net,
                "breakdown": element.breakdown,
                "bill": {
                    "id": element.bill.id,
                    "batch": {
                        "id": element.bill.batch.id,
                        "reference": element.bill.batch.reference,
                    },
                },
            }
        )

        parts["gbp"] += element.net
        vals["actual_net_gbp"] += float(element.net)

        for k, v in element.bd.items():
            if isinstance(v, Decimal):
                v = float(v)

            if isinstance(v, list):
                v = set(v)

            try:
                if isinstance(v, set):
                    parts[k].update(v)
                else:
                    parts[k] += v
            except KeyError:
                parts[k] = v
            except TypeError as detail:
                raise BadRequest(
                    f"For key {k} in {element.bd} the value {v} can't be added to "
                    f"the existing value {parts[k]}. {detail}"
                )

    first_era = None
    for era in sess.scalars(
        select(Era)
        .where(
            Era.supply == supply,
            Era.start_date <= period_finish,
            or_(Era.finish_date == null(), Era.finish_date >= period_start),
        )
        .order_by(Era.start_date)
        .distinct()
        .options(
            joinedload(Era.channels),
            joinedload(Era.cop),
            joinedload(Era.dc_contract),
            joinedload(Era.exp_llfc),
            joinedload(Era.exp_llfc).joinedload(Llfc.voltage_level),
            joinedload(Era.exp_supplier_contract),
            joinedload(Era.imp_llfc),
            joinedload(Era.imp_llfc).joinedload(Llfc.voltage_level),
            joinedload(Era.imp_supplier_contract),
            joinedload(Era.mop_contract),
            joinedload(Era.mtc_participant).joinedload(MtcParticipant.meter_type),
            joinedload(Era.pc),
            joinedload(Era.supply).joinedload(Supply.dno),
            joinedload(Era.supply).joinedload(Supply.gsp_group),
            joinedload(Era.supply).joinedload(Supply.source),
        )
    ).unique():
        first_era = era
        chunk_start = hh_max(period_start, era.start_date)
        chunk_finish = hh_min(period_finish, era.finish_date)

        if contract not in (
            era.mop_contract,
            era.dc_contract,
            era.imp_supplier_contract,
            era.exp_supplier_contract,
        ):
            virtual_bill["problem"] += (
                f"From {date_format(chunk_start)} to {date_format(chunk_finish)} "
                f"the contract of the era doesn't match the contract of the bill."
            )
            continue

        if contract.market_role.code == "X":
            polarity = contract != era.exp_supplier_contract
        else:
            polarity = era.imp_supplier_contract is not None

        data_source = SupplySource(
            sess,
            chunk_start,
            chunk_finish,
            forecast_date,
            era,
            polarity,
            caches,
            bill=True,
        )
        vbf(data_source)

        match market_role_code:
            case "X":
                vb = data_source.supplier_bill
            case "C":
                vb = data_source.dc_bill
            case "M":
                vb = data_source.mop_bill
            case _:
                raise BadRequest(f"Odd market role {market_role_code}")

        for k, v in vb.items():
            if k.endswith("-gbp") and k not in ("net-gbp", "vat-gbp", "gross-gbp"):
                vel_name = k[:-4]
                try:
                    vel = vels[vel_name]
                except KeyError:
                    vel = vels[vel_name] = {"parts": {}, "elements": []}

                vals["virtual_net_gbp"] += v

        for k, v in vb.items():
            if k == "problem":
                virtual_bill["problem"] += v
            else:
                for vel_name in sorted(vels.keys(), key=len, reverse=True):
                    pref = f"{vel_name}-"
                    if k.startswith(pref):
                        vel = vels[vel_name]["parts"]
                        vel_k = k[len(pref) :]
                        try:
                            if isinstance(vel[vel_k], set):
                                vel[vel_k].update(v)
                            else:
                                vel[vel_k] += v
                        except KeyError:
                            vel[vel_k] = v
                        except TypeError as detail:
                            raise BadRequest(f"For key {vel_k} and value {v}. {detail}")

                        break
    for typ, els in (("virtual", vels), ("actual", actual_elems)):
        for el_k, el in els.items():
            try:
                val_elem = val_elems[el_k]
            except KeyError:
                val_elem = val_elems[el_k] = {}

            for k, v in el["parts"].items():
                try:
                    val_parts = val_elem["parts"]
                except KeyError:
                    val_parts = val_elem["parts"] = {}

                try:
                    val_part = val_parts[k]
                except KeyError:
                    val_part = val_parts[k] = {}

                val_part[typ] = v

            for el in el["elements"]:
                try:
                    elements = val_elem[f"{typ}_elements"]
                except KeyError:
                    elements = val_elem[f"{typ}_elements"] = []

                elements.append(el)

    for elname, val_elem in val_elems.items():
        for part_name, part in val_elem["parts"].items():
            if part_name == "gbp":
                virtual_part = round(part.get("virtual", 0), 2)
                actual_part = part.get("actual", 0)
            else:
                virtual_part = part.get("virtual")
                actual_part = part.get("actual")

            if isinstance(virtual_part, set) and len(virtual_part) == 1:
                virtual_part = next(iter(virtual_part))
            if isinstance(actual_part, set) and len(actual_part) == 1:
                actual_part = next(iter(actual_part))

            if virtual_part is None or actual_part is None:
                diff = None
                passed = "❔"
            elif isinstance(virtual_part, Number) and isinstance(actual_part, Number):
                diff = float(actual_part) - float(virtual_part)
                passed = "✅" if diff == 0 else "❌"
            else:
                diff = None
                passed = "✅" if virtual_part == actual_part else "❌"

            part["difference"] = diff
            part["passed"] = passed

    if first_era is None:
        vals["problem"] += "No eras for this period of the supply. "
        first_era = supply.find_last_era(sess)

    site = first_era.get_physical_site(sess)

    vals["site_id"] = site.id
    vals["site_code"] = site.code
    vals["site_name"] = site.name
    vals["imp_mpan_core"] = first_era.imp_mpan_core
    vals["exp_mpan_core"] = first_era.exp_mpan_core
    vals["difference_net_gbp"] = vals["actual_net_gbp"] - vals["virtual_net_gbp"]
    vals["problem"] += virtual_bill["problem"]

    return vals


def _process_supply(sess, caches, supply_id, bill_ids, forecast_date, contract, vbf):
    gaps = {}
    bill_statuses = {}
    supply = Supply.get_by_id(sess, supply_id)
    market_role_code = contract.market_role.code  # noqa: F841

    # Find seed gaps
    while len(bill_ids) > 0:
        bill_id = list(sorted(bill_ids))[0]
        bill_ids.remove(bill_id)
        bill = Bill.get_by_id(sess, bill_id)
        if _get_bill_status(sess, bill_statuses, bill) is not None:
            _add_gap(caches, gaps, bill.start_date, bill.finish_date)
            for element in sess.scalars(
                select(Element)
                .join(Bill)
                .join(Batch)
                .where(
                    Batch.contract == contract,
                    Bill.supply == supply,
                    Bill.start_date <= bill.finish_date,
                    Bill.finish_date >= bill.start_date,
                )
            ):
                _add_gap(caches, gaps, element.start_date, element.finish_date)

    # Find enlarged gaps
    enlarged = True
    while enlarged:
        enlarged = False
        for gap_start, gap_finish in find_gaps(gaps):
            for element in sess.scalars(
                select(Element)
                .join(Bill)
                .join(Batch)
                .where(
                    Bill.supply == supply,
                    Bill.start_date <= gap_finish,
                    Bill.finish_date >= gap_start,
                    Batch.contract == contract,
                )
            ):
                if _add_gap(caches, gaps, element.start_date, element.finish_date):
                    enlarged = True

    for period_start, period_finish in find_gaps(gaps):
        yield _process_period(
            sess,
            caches,
            supply,
            contract,
            bill_statuses,
            forecast_date,
            vbf,
            period_start,
            period_finish,
        )

        # Avoid long-running transactions
        sess.rollback()



from datetime import datetime as Datetime

from pytz import utc

from chellow.utils import (
    PropDict,
    c_months_u,
    ct_datetime,
    date_format,
    make_val,
    parse_date,
    parse_mpan_core,
    to_utc,
    u_months_u,
    utc_datetime,
)


def test_make_val():
    v = {0, "hello"}
    make_val(v)


def test_PropDict_create():
    location = "loc"
    props = {0: [{0: 1}]}
    PropDict(location, props)


def test_PropDict_get():
    assert PropDict("cont " + str(Datetime(2017, 1, 1)), {}, []).get("akey") is None


def test_PropDict():
    assert PropDict("", {"*": 5})[1] == 5


def test_PropDict_nested():
    assert PropDict("", {1: {"*": 5}})[1][1] == 5


def test_to_utc():
    dt_utc = to_utc(ct_datetime(2014, 9, 6, 1))
    assert dt_utc == Datetime(2014, 9, 6, 0, 0, tzinfo=utc)


def test_c_months_u():
    finish_year, finish_month = 2009, 3
    start, _ = next(
        c_months_u(finish_year=finish_year, finish_month=finish_month, months=1)
    )
    assert start == to_utc(ct_datetime(finish_year, finish_month))


def test_c_months_u_start_finish():
    start_year, start_month, finish_year, finish_month = 2009, 3, 2009, 4
    month_list = list(
        c_months_u(
            start_year=start_year,
            start_month=start_month,
            finish_year=finish_year,
            finish_month=finish_month,
        )
    )
    assert month_list == [
        (to_utc(ct_datetime(2009, 3)), to_utc(ct_datetime(2009, 3, 31, 23, 30))),
        (to_utc(ct_datetime(2009, 4)), to_utc(ct_datetime(2009, 4, 30, 23, 30))),
    ]


def test_u_months_u_start_none():
    start_year, start_month = 2009, 3
    month_1 = next(
        u_months_u(start_year=start_year, start_month=start_month, months=None)
    )
    assert month_1 == (utc_datetime(2009, 3), utc_datetime(2009, 3, 31, 23, 30))


def test_date_format_ct():
    dt = utc_datetime(2019, 6, 30)
    actual = date_format(dt)
    assert actual == ("2019-06-30 01:00")


def test_date_format_hh():
    dt = utc_datetime(2019, 6, 30)
    actual = date_format(dt, with_hh=True)
    assert actual == ("2019-06-30 01:00", 3)


def test_date_format_hh_46():
    dt = to_utc(ct_datetime(2019, 3, 31, 23, 30))
    actual = date_format(dt, with_hh=True)
    assert actual == ("2019-03-31 23:30", 46)


def test_date_format_hh_50():
    dt = to_utc(ct_datetime(2019, 10, 27, 23, 30))
    actual = date_format(dt, with_hh=True)
    assert actual == ("2019-10-27 23:30", 50)


def test_date_format_none():
    dt = None
    actual = date_format(dt)
    assert actual == "ongoing"


def test_parse_date():
    actual = parse_date("2019-01-01 00:00Z")
    assert actual == utc_datetime(2019, 1, 1)


def test_parse_mpan_core():
    actual = parse_mpan_core("20\u00a019010100002")
    assert actual == "20 1901 0100 002"



import atexit
import collections
import threading
import traceback

from dateutil.relativedelta import relativedelta

from sqlalchemy import false, select

from chellow.models import ReportRun, Session
from chellow.utils import utc_datetime_now

rrun_deleter = None


def startup():
    global rrun_deleter

    rrun_deleter = ReportRunDeleter()
    rrun_deleter.start()


class ReportRunDeleter(threading.Thread):
    def __init__(self):
        super(ReportRunDeleter, self).__init__(name="ReportRun Deleter")
        self.messages = collections.deque(maxlen=1000)
        self.stopped = threading.Event()
        self.going = threading.Event()
        self.global_alert = None

    def stop(self):
        self.stopped.set()
        self.going.set()
        self.join()

    def go(self):
        self.going.set()

    def log(self, message):
        self.messages.appendleft(
            utc_datetime_now().strftime("%Y-%m-%d %H:%M:%S") + " - " + message
        )

    def run(self):
        while not self.stopped.isSet():
            try:
                with Session() as sess:
                    now = utc_datetime_now()
                    cutoff_date = now - relativedelta(years=1)
                    for report_run in sess.scalars(
                        select(ReportRun).where(
                            ReportRun.date_created < cutoff_date,
                            ReportRun.data["keep"].as_boolean() == false(),
                        )
                    ):
                        report_run.delete(sess)
                        sess.commit()
                        self.log("Deleted report run")
            except BaseException:
                self.log("Outer problem " + traceback.format_exc())
                self.global_alert = (
                    "There's a problem with a "
                    "<a href='/report_runs'>Report Runs</a>."
                )
            finally:
                self.going.clear()
                self.log("Finished deleting Report Runs.")

            self.going.wait(24 * 60 * 60)
            self.going.clear()


def get_importer():
    return rrun_deleter


@atexit.register
def shutdown():
    if rrun_deleter is not None:
        rrun_deleter.stop()



from sqlalchemy import select

from chellow.e.neso import csv_latest, parse_date
from chellow.models import Contract, RateScript
from chellow.utils import ct_datetime, to_utc

MAXIMA = [
    {
        "maximum": 15,
        "start": to_utc(ct_datetime(2020, 6, 25)),
        "finish": to_utc(ct_datetime(2020, 8, 31, 23, 30)),
        "reference": "CMP345",
    },
    {
        "maximum": 20,
        "start": to_utc(ct_datetime(2022, 1, 17)),
        "finish": to_utc(ct_datetime(2022, 3, 31, 23, 30)),
        "reference": "CMP381",
    },
    {
        "maximum": 40,
        "start": to_utc(ct_datetime(2022, 10, 6)),
        "finish": to_utc(ct_datetime(2023, 3, 31, 23, 30)),
        "reference": "CMP395",
    },
]


def hh(data_source, run="RF"):
    try:
        bsuos_cache = data_source.caches["bsuos"][run]
    except KeyError:
        try:
            top_cache = data_source.caches["bsuos"]
        except KeyError:
            top_cache = data_source.caches["bsuos"] = {}

        try:
            bsuos_cache = top_cache[run]
        except KeyError:
            bsuos_cache = top_cache[run] = {}

    for h in data_source.hh_data:
        try:
            h["bsuos-rate"] = bsuos_rate = bsuos_cache[h["start-date"]]
        except KeyError:
            h_start = h["start-date"]

            rates = data_source.non_core_rate("bsuos", h_start)

            if h_start >= to_utc(ct_datetime(2023, 4, 1)):
                bsuos_price = float(rates["record"]["Fixed Tariff £/MWh"])
            else:
                bsuos_rates = rates["rates_gbp_per_mwh"]

                maxi = None
                for max_dict in MAXIMA:
                    if max_dict["start"] <= h_start <= max_dict["finish"]:
                        maxi = max_dict["maximum"]
                        break

                try:
                    bsuos_price_dict = bsuos_rates[key_format(h_start)]
                    bsuos_price = _find_price(run, bsuos_price_dict, maxi)
                except KeyError:
                    ds = bsuos_rates.values()
                    bsuos_price = sum(_find_price(run, d, maxi) for d in ds) / len(ds)

            h["bsuos-rate"] = bsuos_rate = bsuos_cache[h_start] = bsuos_price / 1000

        h["bsuos-kwh"] = h["nbp-kwh"]
        h["bsuos-gbp"] = h["nbp-kwh"] * bsuos_rate


def _find_price(run, prices, maxi):
    try:
        price = prices[run]
    except KeyError:
        try:
            price = prices["RF"]
        except KeyError:
            try:
                price = prices["SF"]
            except KeyError:
                price = prices["II"]

    float_price = float(price)
    return float_price if maxi is None else min(float_price, maxi)


def key_format(dt):
    return dt.strftime("%d %H:%M")


CONTRACT_NAME = "bsuos"


def neso_import(sess, log, set_progress):
    log("Starting to check for new BSUoS rates")

    CONTRACT_NAME = "bsuos"
    contract = Contract.find_non_core_by_name(sess, CONTRACT_NAME)
    if contract is None:
        contract = Contract.insert_non_core(
            sess, CONTRACT_NAME, "", {}, to_utc(ct_datetime(1996, 4, 1)), None, {}
        )
    state = contract.make_state()
    last_import_date = state.get("last_import_date")

    for record in csv_latest("bsuos-fixed-tariffs", last_import_date):
        #   "Publication": "Final",
        #   "Fixed Tariff Title": "Fixed Tariff 1",
        #   "Published Date": "2023-01-31",
        #   "Fixed Tariff Start Date": "2023-04-01",
        #   "Fixed Tariff End Date": "2023-09-30",
        #   "Fixed Tariff £/MWh": "13.41",

        tariff_start = parse_date(record["Fixed Tariff Start Date"])
        rs = sess.execute(
            select(RateScript).where(
                RateScript.contract == contract,
                RateScript.start_date == tariff_start,
            )
        ).scalar_one_or_none()
        if rs is None:
            rs = contract.insert_rate_script(sess, tariff_start, {})

        rs_script = rs.make_script()
        try:
            rs_record = rs_script["record"]
        except KeyError:
            rs_record = rs_script["record"] = {}

        record_published_date = parse_date(record["Published Date"])

        if (
            "Published Date" not in rs_record
            or parse_date(rs_record["Published Date"]) < record_published_date
        ):
            rs_script["record"] = record
            rs.update(rs_script)
            sess.commit()

    log("Finished Checking for BSUoS Tariffs")
    sess.commit()



import csv
import sys
import threading
import traceback

from flask import g, redirect

from sqlalchemy import or_
from sqlalchemy.sql.expression import null

from werkzeug.exceptions import BadRequest

from chellow.dloads import open_file
from chellow.e.computer import SupplySource, contract_func, forecast_date
from chellow.models import Contract, Era, Session, User
from chellow.utils import csv_make_val, date_format, hh_max, hh_min, req_date, req_int


def content(user_id, start_date, finish_date, contract_id):
    caches = {}
    supply_source = f = None
    try:
        with Session() as sess:
            user = User.get_by_id(sess, user_id)
            f = open_file("mop_virtual_bills.csv", user, mode="w", newline="")
            writer = csv.writer(f, lineterminator="\n")
            contract = Contract.get_mop_by_id(sess, contract_id)

            f_date = forecast_date()
            header_titles = [
                "imp_mpan_core",
                "exp_mpan_core",
                "start_date",
                "finish_date",
                "energisation_status",
                "gsp_group",
                "dno",
                "era_start",
                "pc",
                "meter_type",
                "site_code",
                "imp_is_substation",
                "imp_llfc_code",
                "imp_llfc_description",
                "exp_is_substation",
                "exp_llfc_code",
                "exp_llfc_description",
            ]

            bill_titles = contract_func(caches, contract, "virtual_bill_titles")()
            titles = header_titles + bill_titles
            writer.writerow(titles)
            vb_func = contract_func(caches, contract, "virtual_bill")

            for era in (
                sess.query(Era)
                .filter(
                    or_(Era.finish_date == null(), Era.finish_date >= start_date),
                    Era.start_date <= finish_date,
                    Era.mop_contract == contract,
                )
                .order_by(Era.imp_mpan_core, Era.exp_mpan_core, Era.start_date)
            ):
                chunk_start = hh_max(era.start_date, start_date)
                chunk_finish = hh_min(era.finish_date, finish_date)
                is_import = era.imp_mpan_core is not None

                supply_source = SupplySource(
                    sess, chunk_start, chunk_finish, f_date, era, is_import, caches
                )

                if is_import:
                    imp_is_substation = supply_source.is_substation
                    imp_llfc_code = supply_source.llfc_code
                    imp_llfc_description = supply_source.llfc.description
                    exp_is_substation = exp_llfc_code = exp_llfc_description = None
                else:
                    exp_is_substation = supply_source.is_substation
                    exp_llfc_code = supply_source.llfc_code
                    exp_llfc_description = supply_source.llfc.description
                    imp_is_substation = imp_llfc_code = imp_llfc_description = None

                out = {
                    "imp_mpan_core": era.imp_mpan_core,
                    "exp_mpan_core": era.exp_mpan_core,
                    "start_date": chunk_start,
                    "finish_date": chunk_finish,
                    "energisation_status": supply_source.energisation_status_code,
                    "gsp_group": supply_source.gsp_group_code,
                    "dno": supply_source.dno_code,
                    "era_start": era.start_date,
                    "pc": supply_source.pc_code,
                    "meter_type": supply_source.meter_type_code,
                    "site_code": era.get_physical_site(sess).code,
                    "imp_is_substation": imp_is_substation,
                    "imp_llfc_code": imp_llfc_code,
                    "imp_llfc_description": imp_llfc_description,
                    "exp_is_substation": exp_is_substation,
                    "exp_llfc_code": exp_llfc_code,
                    "exp_llfc_description": exp_llfc_description,
                }
                vb_func(supply_source)
                bill = supply_source.mop_bill
                for title in bill_titles:
                    if title in bill:
                        out[title] = bill[title]
                writer.writerow(csv_make_val(out.get(t)) for t in titles)

                sess.rollback()  # Avoid long-running transactions
    except BadRequest as e:
        msg = "Problem "
        if supply_source is not None:
            msg += (
                f"with supply {supply_source.mpan_core} starting at "
                f"{date_format(supply_source.start_date)} "
            )
        msg += str(e)
        sys.stderr.write(msg)
        writer.writerow([msg])
    except BaseException:
        msg = traceback.format_exc()
        sys.stderr.write(msg)
        writer.writerow([msg])
    finally:
        if f is not None:
            f.close()


def do_get(sess):
    start_date = req_date("start")
    finish_date = req_date("finish")
    contract_id = req_int("mop_contract_id")

    args = g.user.id, start_date, finish_date, contract_id
    threading.Thread(target=content, args=args).start()
    return redirect("/downloads", 303)



from decimal import Decimal


from werkzeug.exceptions import BadRequest

from chellow.edi_lib import (
    parse_edi,
    to_date,
    to_decimal,
    to_finish_date,
)
from chellow.utils import HH, parse_mpan_core

read_type_map = {
    "00": "N",
    "09": "N3",
    "04": "C",
    "02": "E",
    "11": "E3",
    "01": "EM",
    "03": "W",
    "06": "X",
    "05": "CP",
    "12": "IF",
}


TCOD_MAP = {
    "140114": ("reconciliation", None, None),
    "255204": ("meter-rental", "rate", "days"),
    "345065": ("op-weekend", "rate", "kwh"),
    "350293": ("capacity", "rate", "kwh"),
    "425779": ("ro", "rate", "kwh"),
    "534342": ("reconciliation", None, None),
    "583174": ("meter-rental", "rate", "days"),
    "584867": ("aahedc", "rate", "kwh"),
    "946827": ("meter-rental", "rate", "days"),
    "989534": ("bsuos", "rate", "kwh"),
    "117220": ("capacity", "rate", "kwh"),
    "579387": ("capacity", "rate", "kwh"),
    "558147": ("capacity", "rate", "kwh"),
    "030025": ("ccl", "rate", "kwh"),
    "066540": ("ccl", "rate", "kwh"),
    "154164": ("cfd-fit", "rate", "kwh"),
    "281170": ("cfd-fit", "rate", "fit-kwh"),
    "342094": ("cfd-fit", "rate", "kwh"),
    "378809": ("cfd-fit", "rate", "kwh"),
    "574015": ("cfd-fit", "rate", "kwh"),
    "810016": ("cfd-fit", "rate", "kwh"),
    "839829": ("cfd-fit", "rate", "kwh"),
    "649282": ("cfd-fit", "rate", "kwh"),
    "068476": ("day", "rate", "kwh"),
    "133186": ("nrg", "rate", "kwh"),
    "400434": ("day", "rate", "kwh"),
    "219182": ("duos-availability", "rate", "kva"),
    "144424": ("duos-excess-availability", "rate", "kva"),
    "301541": ("duos-fixed", None, None),
    "099335": ("duos-fixed", None, None),
    "873562": ("duos-fixed", None, None),
    "986159": ("duos-fixed", "rate", "days"),
    "838286": ("duos-reactive", "rate", "kvarh"),
    "242643": ("duos-fixed", "rate", "days"),
    "257304": ("duos-amber", "rate", "kwh"),
    "661440": ("duos-amber", "rate", "kwh"),
    "257305": ("duos-green", "rate", "kwh"),
    "661441": ("duos-green", "rate", "kwh"),
    "257303": ("duos-red", "rate", "kwh"),
    "661439": ("duos-red", "rate", "kwh"),
    "504364": ("ebrs", None, "kwh"),
    "563023": ("ebrs", None, "kwh"),
    "823408": ("ebrs", None, "kwh"),
    "871593": ("ebrs", "rate", "kwh"),
    "873894": ("ebrs", "rate", "kwh"),
    "309707": ("fit", "rate", "kwh"),
    "310129": ("meter-rental", None, None),
    "452415": ("meter-rental", None, None),
    "371265": ("meter-rental", None, None),
    "544936": ("meter-rental", "rate", "days"),
    "265091": ("night", "rate", "kwh"),
    "483457": ("peak", "rate", "kwh"),
    "975901": ("peak-shoulder", "rate", "kwh"),
    "994483": ("reconciliation", None, None),
    "637176": ("reconciliation", None, None),
    "913821": ("reconciliation", None, None),
    "307660": ("ro", "rate", "kwh"),
    "364252": ("ro", "rate", "kwh"),
    "378246": ("ro", "rate", "kwh"),
    "708848": ("ro", None, None),
    "632209": ("summer-night", "rate", "kwh"),
    "663682": ("summer-weekday", "rate", "kwh"),
    "299992": ("summer-weekend", "rate", "kwh"),
    "211000": ("tnuos", "rate", "days"),
    "790618": ("tnuos", None, None),
    "447769": ("triad", "rate", "kw"),
    "647721": ("triad", "rate", "kw"),
    "276631": ("triad", "rate", "kw"),
    "220894": ("winter-night", "rate", "kwh"),
    "264929": ("winter-weekday", "rate", "kwh"),
    "638187": ("winter-weekend", "rate", "kwh"),
    "700285": ("standing", "rate", "days"),
}

TPR_LOOKUP = {
    "Day": "00043",
    "Off Peak / Weekends": "00210",
    "Night": "00210",
    "Default Rate": "00043",
    "Single": "00210",
}


def _process_BCD(elements, headers):
    issue_date = to_date(elements["IVDT"][0])
    reference = elements["INVN"][0]
    bill_type_code = elements["BTCD"][0]
    sumo = elements["SUMO"]

    headers["issue_date"] = issue_date
    headers["start_date"] = to_date(sumo[0])
    headers["finish_date"] = to_date(sumo[1]) - HH
    headers["bill_type_code"] = bill_type_code
    headers["reference"] = reference
    headers["elements"] = []
    headers["reads"] = []
    headers["breakdown"] = {}
    headers["kwh"] = Decimal("0")


def _process_BTL(elements, headers):
    uvlt = elements["UVLT"]
    utva = elements["UTVA"]
    tbtl = elements["TBTL"]

    return {
        "bill_type_code": headers["bill_type_code"],
        "reference": headers["reference"],
        "issue_date": headers["issue_date"],
        "mpan_core": headers["mpan_core"],
        "account": headers["account"],
        "start_date": headers["start_date"],
        "finish_date": headers["finish_date"],
        "kwh": headers["kwh"],
        "net": Decimal("0.00") + to_decimal(uvlt) / Decimal("100"),
        "vat": Decimal("0.00") + to_decimal(utva) / Decimal("100"),
        "gross": Decimal("0.00") + to_decimal(tbtl) / Decimal("100"),
        "breakdown": headers["breakdown"],
        "reads": headers["reads"],
        "elements": headers["elements"],
    }


def _process_CCD1(elements, headers):
    tcod = elements["TCOD"]
    pres_read_date = to_finish_date(elements["PRDT"][0])

    prev_read_date = to_finish_date(elements["PVDT"][0])

    m = elements["MLOC"][0]
    mpan = " ".join((m[13:15], m[15:18], m[18:], m[:2], m[2:6], m[6:10], m[10:13]))

    prrd = elements["PRRD"]
    if len(prrd) < 4:
        return
    pres_read_type = read_type_map[prrd[1]]
    prev_read_type = read_type_map[prrd[3]]

    coefficient = Decimal(elements["ADJF"][1]) / Decimal(100000)
    pres_reading_value = Decimal(prrd[0])
    prev_reading_value = Decimal(prrd[2])
    msn = elements["MTNR"][0]
    tpr_code = elements["TMOD"][0]
    if tpr_code == "kW":
        units = "kW"
        tpr_code = None
    elif tpr_code == "kVA":
        units = "kVA"
        tpr_code = None
    else:
        units = "kWh"
        tpr_code = TPR_LOOKUP[tcod[1]]

    headers["reads"].append(
        {
            "msn": msn,
            "mpan": mpan,
            "coefficient": coefficient,
            "units": units,
            "tpr_code": tpr_code,
            "prev_date": prev_read_date,
            "prev_value": prev_reading_value,
            "prev_type_code": prev_read_type,
            "pres_date": pres_read_date,
            "pres_value": pres_reading_value,
            "pres_type_code": pres_read_type,
        }
    )


def _process_CCD2(elements, headers):
    breakdown = {}

    element_code = elements["TCOD"][0]
    try:
        eln_name, eln_rate, eln_cons = TCOD_MAP[element_code]
    except KeyError:
        raise BadRequest(f"Can't find the element code {element_code} in the TCOD_MAP.")

    cons = elements["CONS"]
    if eln_cons is not None and len(cons[0]) > 0:
        el_cons = to_decimal(cons) / Decimal("1000")
        if eln_name == "duos-availability":
            breakdown[eln_cons] = {el_cons}
        else:
            breakdown[eln_cons] = el_cons
            if eln_name == "ro":
                headers["kwh"] += el_cons

    if eln_rate is not None:
        rate = to_decimal(elements["BPRI"]) / Decimal("100000")
        breakdown[eln_rate] = {rate}

    net = Decimal("0.00")
    if "CTOT" in elements:
        net += to_decimal(elements["CTOT"]) / Decimal("100")

    headers["elements"].append(
        {
            "name": eln_name,
            "start_date": to_date(elements["CSDT"][0]),
            "finish_date": to_date(elements["CEDT"][0]) - HH,
            "net": net,
            "breakdown": breakdown,
        }
    )


def _process_CCD3(elements, headers):
    breakdown = {}

    element_code = elements["TCOD"][0]
    try:
        eln_name, eln_rate, eln_cons = TCOD_MAP[element_code]
    except KeyError:
        raise BadRequest(f"Can't find the element code {element_code} in the TCOD_MAP.")

    cons = elements["CONS"]
    if eln_cons is not None and len(cons[0]) > 0:
        el_cons = to_decimal(cons) / Decimal("1000")
        breakdown[eln_cons] = el_cons
        if eln_name == "ro":
            headers["kwh"] += el_cons

    if eln_rate is not None:
        rate = to_decimal(elements["BPRI"]) / Decimal("100000")
        breakdown[eln_rate] = {rate}

    net = Decimal("0.00")
    if "CTOT" in elements:
        net += to_decimal(elements["CTOT"]) / Decimal("100")

    headers["elements"].append(
        {
            "name": eln_name,
            "start_date": to_date(elements["CSDT"][0]),
            "finish_date": to_date(elements["CEDT"][0]) - HH,
            "net": net,
            "breakdown": breakdown,
        }
    )


def _process_CCD4(elements, headers):
    breakdown = {}

    element_code = elements["TCOD"][0]

    try:
        eln_name, eln_rate, eln_cons = TCOD_MAP[element_code]
    except KeyError:
        raise BadRequest(f"Can't find the element code {element_code} in the TCOD_MAP.")

    cons = elements["CONS"]
    if eln_cons is not None and len(cons[0]) > 0:
        el_cons = to_decimal(cons, "1000")
        breakdown[eln_cons] = el_cons
        if eln_name == "ro":
            headers["kwh"] += el_cons

    if eln_rate is not None:
        rate = to_decimal(elements["BPRI"], "100000")
        breakdown[eln_rate] = [rate]

    net = Decimal("0.00")
    if "CTOT" in elements:
        net += to_decimal(elements["CTOT"], "100")

    headers["elements"].append(
        {
            "name": eln_name,
            "start_date": to_date(elements["CSDT"][0]),
            "finish_date": to_date(elements["CEDT"][0]) - HH,
            "net": net,
            "breakdown": breakdown,
        }
    )


def _process_CDT(elements, headers):
    customer_id = elements["CIDN"][0]
    headers["customer_number"] = customer_id


def _process_CLO(elements, headers):
    cloc = elements["CLOC"]
    headers["account"] = cloc[2]


def _process_END(elements, headers):
    pass


def _process_MAN(elements, headers):
    madn = elements["MADN"]
    headers["mpan_core"] = parse_mpan_core("".join(madn[:3]))


def _process_MHD(elements, headers):
    message_type = elements["TYPE"][0]
    if message_type == "UTLBIL":
        keep_keys = {"customer_number"}
        keep = {k: headers[k] for k in keep_keys}
        headers.clear()
        headers.update(keep)


def _process_MTR(elements, headers):
    pass


def _process_VAT(elements, headers):
    vat = Decimal("0.00") + to_decimal(elements["UVTT"]) / Decimal("100")
    vat_percentage = to_decimal(elements["VATP"]) / Decimal("1000")
    vat_net = Decimal("0.00") + to_decimal(elements["UVLA"]) / Decimal("100")

    breakdown = headers["breakdown"]

    try:
        vat_bd = breakdown["vat"]
    except KeyError:
        vat_bd = breakdown["vat"] = {}

    vat_bd[vat_percentage] = {"vat": vat, "net": vat_net}


def _process_NOOP(elements, headers):
    pass


CODE_FUNCS = {
    "BCD": _process_BCD,
    "BTL": _process_BTL,
    "CCD1": _process_CCD1,
    "CCD2": _process_CCD2,
    "CCD3": _process_CCD3,
    "CCD4": _process_CCD4,
    "CDT": _process_CDT,
    "CLO": _process_CLO,
    "DNA": _process_NOOP,
    "END": _process_END,
    "FIL": _process_NOOP,
    "MAN": _process_MAN,
    "MHD": _process_MHD,
    "MTR": _process_MTR,
    "SDT": _process_NOOP,
    "STX": _process_NOOP,
    "TYP": _process_NOOP,
    "TTL": _process_NOOP,
    "VAT": _process_VAT,
    "VTS": _process_NOOP,
}


class Parser:
    def __init__(self, f):
        self.edi_str = str(f.read(), "utf-8", errors="ignore")
        self.line_number = None

    def make_raw_bills(self):
        bills = []
        headers = {}
        bill = None
        lines = []
        for self.line_number, line, seg_name, elements in parse_edi(self.edi_str):
            if seg_name == "MHD":
                lines = []
            lines.append(line)

            try:
                func = CODE_FUNCS[seg_name]
            except KeyError:
                raise BadRequest(f"Code {seg_name} not recognized.")

            try:
                bill = func(elements, headers)
            except BadRequest as e:
                raise BadRequest(
                    f"{e.description} on line {self.line_number} line {line} "
                    f"seg_name {seg_name} elements {elements}"
                )
            except BaseException as e:
                raise BadRequest(
                    f"{e} on line {self.line_number} line {line} "
                    f"seg_name {seg_name} elements {elements}"
                ) from e

            if bill is not None:
                bill["breakdown"]["raw-lines"] = lines
                bills.append(bill)
            if seg_name == "MTR":
                lines = []

        return bills



import csv
import threading
import traceback

from flask import g, redirect, request

from sqlalchemy import and_, or_, select
from sqlalchemy.orm import joinedload

from werkzeug.exceptions import BadRequest

from chellow.dloads import open_file
from chellow.models import (
    Batch,
    Bill,
    BillType,
    Era,
    RSession,
    RegisterRead,
    Supply,
    User,
)
from chellow.utils import c_months_u, csv_make_val, req_int


def content(year, month, months, supply_id, user_id):
    f = None
    try:
        with RSession() as sess:
            user = User.get_by_id(sess, user_id)

            f = open_file("register_reads.csv", user, mode="w", newline="")
            w = csv.writer(f, lineterminator="\n")
            titles = (
                "duration_start",
                "duration_finish",
                "site_code",
                "site_name",
                "supply_id",
                "imp_mpan_core",
                "exp_mpan_core",
                "batch_reference",
                "bill_id",
                "bill_reference",
                "bill_issue_date",
                "bill_type",
                "register_read_id",
                "tpr",
                "coefficient",
                "prev_read_date",
                "prev_read_value",
                "prev_read_type",
                "pres_read_date",
                "pres_read_value",
                "pres_read_type",
            )
            w.writerow(titles)

            month_pairs = list(
                c_months_u(finish_year=year, finish_month=month, months=months)
            )
            start_date, finish_date = month_pairs[0][0], month_pairs[-1][-1]

            supplies = (
                select(Supply)
                .join(Bill)
                .join(RegisterRead)
                .where(
                    or_(
                        and_(
                            RegisterRead.present_date >= start_date,
                            RegisterRead.present_date <= finish_date,
                        ),
                        and_(
                            RegisterRead.previous_date >= start_date,
                            RegisterRead.previous_date <= finish_date,
                        ),
                    )
                )
                .order_by(Supply.id)
                .distinct()
            )

            if supply_id is not None:
                supply = Supply.get_by_id(sess, supply_id)
                supplies = supplies.where(Supply.id == supply.id)

            for supply in sess.scalars(supplies):
                supply_id = supply.id
                for bill, batch, bill_type in sess.execute(
                    select(Bill, Batch, BillType)
                    .join(Batch)
                    .join(BillType)
                    .join(RegisterRead)
                    .filter(
                        Bill.supply == supply,
                        or_(
                            and_(
                                RegisterRead.present_date >= start_date,
                                RegisterRead.present_date <= finish_date,
                            ),
                            and_(
                                RegisterRead.previous_date >= start_date,
                                RegisterRead.previous_date <= finish_date,
                            ),
                        ),
                    )
                ):
                    era = supply.find_era_at(sess, bill.start_date)
                    if era is None:
                        eras = (
                            sess.query(Era)
                            .filter(Era.supply == supply)
                            .order_by(Era.start_date)
                            .all()
                        )
                        if bill.start_date < eras[0].start_date:
                            era = eras[0]
                        else:
                            era = eras[-1]

                    site = era.get_physical_site(sess)

                    for read in (
                        sess.query(RegisterRead)
                        .filter(
                            RegisterRead.bill == bill,
                            or_(
                                and_(
                                    RegisterRead.present_date >= start_date,
                                    RegisterRead.present_date <= finish_date,
                                ),
                                and_(
                                    RegisterRead.previous_date >= start_date,
                                    RegisterRead.previous_date <= finish_date,
                                ),
                            ),
                        )
                        .options(
                            joinedload(RegisterRead.tpr),
                            joinedload(RegisterRead.previous_type),
                            joinedload(RegisterRead.present_type),
                        )
                    ):
                        vals = {
                            "duration_start": start_date,
                            "duration_finish": finish_date,
                            "site_code": site.code,
                            "site_name": site.name,
                            "supply_id": supply_id,
                            "imp_mpan_core": era.imp_mpan_core,
                            "exp_mpan_core": era.exp_mpan_core,
                            "batch_reference": batch.reference,
                            "bill_id": bill.id,
                            "bill_reference": bill.reference,
                            "bill_issue_date": bill.issue_date,
                            "bill_type": bill_type.code,
                            "register_read_id": read.id,
                            "tpr": "md" if read.tpr is None else read.tpr.code,
                            "coefficient": read.coefficient,
                            "prev_read_date": read.previous_date,
                            "prev_read_value": read.previous_value,
                            "prev_read_type": read.previous_type.code,
                            "pres_read_date": read.present_date,
                            "pres_read_value": read.present_value,
                            "pres_read_type": read.present_type.code,
                        }
                        w.writerow(csv_make_val(vals.get(t)) for t in titles)

                    # Avoid a long-running transaction
                    sess.rollback()

    except BadRequest as e:
        w.writerow([e.description])
    except BaseException:
        msg = traceback.format_exc()
        print(msg)
        f.write(msg)
    finally:
        if f is not None:
            f.close()


def do_get(sess):
    year = req_int("end_year")
    month = req_int("end_month")
    months = req_int("months")
    supply_id = req_int("supply_id") if "supply_id" in request.values else None
    args = year, month, months, supply_id, g.user.id
    threading.Thread(target=content, args=args).start()
    return redirect("/downloads", 303)



import csv
from io import StringIO

from utils import match, match_tables

from chellow.models import (
    Contract,
    MarketRole,
    Participant,
    User,
    UserRole,
)
from chellow.reports.report_issues import content
from chellow.utils import ct_datetime, to_utc


def test_do_get_as_csv(mocker, sess, client):
    mock_Thread = mocker.patch("chellow.reports.report_issues.threading.Thread")
    vf = to_utc(ct_datetime(1996, 1, 1))
    participant = Participant.insert(sess, "hhak", "AK Industries")
    market_role_C = MarketRole.insert(sess, "C", "DC")
    participant.insert_party(sess, market_role_C, "Fusion", vf, None, None)
    contract = Contract.insert_dc(sess, "Fusion DC", participant, "", {}, vf, None, {})
    contract.insert_issue(sess, vf, {})
    contract.insert_issue(sess, to_utc(ct_datetime(1997, 1, 1)), {})
    sess.commit()

    query_string = {"contract_id": contract.id, "as_csv": "true"}
    response = client.get("/reports/issues", query_string=query_string)
    match(response, 303)

    args = (1, [2], [], [], [])
    mock_Thread.assert_called_with(target=content, args=args)


def test_content(mocker, sess):
    vf = to_utc(ct_datetime(1996, 1, 1))
    participant = Participant.insert(sess, "hhak", "AK Industries")
    market_role_C = MarketRole.insert(sess, "C", "DC")
    participant.insert_party(sess, market_role_C, "Fusion", vf, None, None)
    contract = Contract.insert_dc(sess, "Fusion DC", participant, "", {}, vf, None, {})
    contract.insert_issue(sess, vf, {})
    contract.insert_issue(sess, to_utc(ct_datetime(1997, 1, 1)), {})
    editor = UserRole.insert(sess, "editor")
    user = User.insert(sess, "admin@example.com", "xxx", editor, None)
    user_id = user.id
    contract_ids = [contract.id]
    sess.commit()

    f = StringIO()
    f.close = mocker.Mock()
    mocker.patch("chellow.reports.report_issues.open_file", return_value=f)
    user_ids = []
    supply_ids = []
    is_opens = []
    content(user_id, contract_ids, user_ids, supply_ids, is_opens)
    f.seek(0)
    actual = list(csv.reader(f))
    expected = [
        [
            "issue_id",
            "contract_role",
            "contract_name",
            "date_created",
            "owner",
            "status",
            "subject",
            "imp_mpan_core",
            "exp_mpan_core",
            "site_code",
            "site_name",
            "latest_entry_timestamp",
            "latest_entry_markdown",
        ],
        [
            "1",
            "C",
            "Fusion DC",
            "1996-01-01 00:00",
            "",
            "open",
            "",
            "",
            "",
            "",
            "",
            "",
            "",
        ],
        [
            "2",
            "C",
            "Fusion DC",
            "1997-01-01 00:00",
            "",
            "open",
            "",
            "",
            "",
            "",
            "",
            "",
            "",
        ],
    ]
    match_tables(expected, actual)



import csv
from datetime import datetime as Datetime
from decimal import Decimal, InvalidOperation
from enum import Enum, auto

from dateutil.relativedelta import relativedelta

from openpyxl import load_workbook
from werkzeug.exceptions import BadRequest


from chellow.utils import HH, ct_datetime, parse_mpan_core, to_ct, to_utc

ELEM_MAP = {
    None: None,
    "Charge - Recurring": {
        None: ("duos-fixed-gbp", "duos-fixed-rate", "duos-fixed-days")
    },
    "Meter - Standard": {
        "Energy Bill Relief Scheme": {None: ["ebrs-gbp", "ebrs-msp-kwh"]},
        "Energy Bill Relief Scheme Discount": {None: ["ebrs-gbp", "ebrs-msp-kwh"]},
        "Energy Bill Discount Scheme": {None: ["ebrs-gbp", "ebrs-msp-kwh"]},
    },
    "Meter - UK Electricity - AAHEDC Pass-Thru": {
        None: ["aahedc-gbp", "aahedc-rate", "aahedc-gsp-kwh"]
    },
    "Meter - UK Electricity - BSUoS Pass-Thru": {
        None: ["bsuos-gbp", "bsuos-rate", "bsuos-nbp-kwh"]
    },
    "Meter - UK Electricity - Capacity Market Pass-Thru": {
        None: ["capacity-gbp", "capacity-rate", "capacity-kwh"],
        "Reverse Capacity Market Estimate": {None: ["capacity-gbp"]},
    },
    "Meter - UK Electricity - CfD FiT Pass-Thru": {
        None: ["cfd-fit-gbp", "cfd-fit-rate", "cfd-fit-nbp-kwh"]
    },
    "Meter - UK Electricity - CCL": {
        None: ["ccl-gbp", "ccl-rate", "ccl-kwh"],
        "CCL": {None: ["ccl-gbp", "ccl-rate", "ccl-kwh"]},
        "Levy Exempt Energy": {None: ["lec-gbp", "lec-rate", "lec-kwh"]},
    },
    "Meter - UK Electricity - DUoS": {
        None: None,
        "DUoS Unit Rate 3": {
            None: ("duos-green-gbp", "duos-green-rate", "duos-green-kwh")
        },
        "DUoS Unit Charge 3": {
            None: ("duos-green-gbp", "duos-green-rate", "duos-green-kwh")
        },
        "DUoS Unit Charge 2": {
            None: ("duos-amber-gbp", "duos-amber-rate", "duos-amber-kwh")
        },
        "DUoS Unit Rate 2": {
            None: ("duos-amber-gbp", "duos-amber-rate", "duos-amber-kwh")
        },
        "DUoS Unit Charge 1": {None: ("duos-red-gbp", "duos-red-kwh", "duos-red-rate")},
        "DUoS Unit Rate 1": {None: ("duos-red-gbp", "duos-red-kwh", "duos-red-rate")},
        "DUoS Standing Charge": {
            None: ("duos-fixed-gbp", "duos-fixed-rate", "duos-fixed-days")
        },
        "DUoS Fixed": {None: ("duos-fixed-gbp", "duos-fixed-rate", "duos-fixed-days")},
        "DUoS Reactive": {
            None: ("duos-reactive-gbp", "duos-reactive-rate", "duos-reactive-kvarh")
        },
    },
    "Meter - UK Electricity - FiT Pass-Thru": {
        None: ("fit-gbp", "fit-rate", "fit-msp-kwh")
    },
    "Pass Thru - UK Electricity Cost Component": {
        None: ("meter-rental-gbp", "meter-rental-rate", "meter-rental-days")
    },
    "Meter - UK Electricity - RO Pass-Thru": {
        None: ("ro-gbp", "ro-rate", "ro-msp-kwh")
    },
    "Meter - UK Electricity - TUoS": {
        None: ("triad-gbp", "triad-rate", "triad-gsp-kw"),
        "TNUoS Fixed": {
            None: ("tnuos-gbp", "tnuos-rate", "tnuos-days"),
        },
    },
    "Meter - UK Electricity - Standard": {
        None: None,
        "Unit Rate": {
            "Summer Weekday": {
                None: (
                    "summer-weekday-gbp",
                    "summer-weekday-rate",
                    "summer-weekday-gsp-kwh",
                )
            },
            "Peak": {None: ("peak-gbp", "peak-rate", "peak-gsp-kwh")},
            "Peak Shoulder": {
                None: (
                    "peak-shoulder-gbp",
                    "peak-shoulder-gsp-kwh",
                    "peak-shoulder-rate",
                )
            },
            "Summer Night": {
                None: ("summer-night-gbp", "summer-night-rate", "summer-night-gsp-kwh")
            },
            "Summer Weekend & Bank Holiday": {
                None: (
                    "summer-weekend-gbp",
                    "summer-weekend-rate",
                    "summer-weekend-gsp-kwh",
                )
            },
            "Night": {None: ("night-gbp", "night-rate", "night-gsp-kwh")},
            "Winter Weekday": {
                None: (
                    "winter-weekday-gbp",
                    "winter-weekday-rate",
                    "winter-weekday-gsp-kwh",
                )
            },
            "Winter Weekend & Bank Holiday": {
                None: (
                    "winter-weekend-gbp",
                    "winter-weekend-rate",
                    "winter-weekend-gsp-kwh",
                )
            },
            "Winter Night": {
                None: ("winter-night-gbp", "winter-night-rate", "winter-night-gsp-kwh")
            },
            "Day": {None: ("day-gbp", "day-rate", "day-gsp-kwh")},
            "Single": {None: ("day-gbp", "day-rate", "day-gsp-kwh")},
            "Off Peak / Weekends": {None: ("day-gbp", "day-rate", "day-gsp-kwh")},
        },
        "Reverse BSUoS in Unit Rate": {
            None: ("bsuos-reverse-gbp", "bsuos-reverse-rate", "bsuos-reverse-nbp-kwh")
        },
    },
    "Meter - UK Gas - CCL": {None: ("ccl-gbp", "ccl-rate", "ccl-kwh")},
}


def _find_names(tree, path):
    if len(path) > 0:
        try:
            return _find_names(tree[path[0]], path[1:])
        except KeyError:
            pass

    return tree[None]


class Title(Enum):
    METER_POINT = auto()
    BILL_PERIOD = auto()
    FROM_DATE = auto()
    TO_DATE = auto()
    BILL_DATE = auto()
    BILL_NUMBER = auto()
    USAGE = auto()
    PRICE = auto()
    AMOUNT = auto()
    PRODUCT_ITEM_NAME = auto()
    RATE_NAME = auto()
    DESCRIPTION = auto()
    PRODUCT_ITEM_CLASS = auto()
    CUSTOMER_NUMBER = auto()


COLUMNS = {
    Title.METER_POINT: ["Meter Point", "MeterPoint"],
    Title.BILL_PERIOD: ["Bill Period", "BillPeriod"],
    Title.FROM_DATE: ["From Date", "FromDate"],
    Title.TO_DATE: ["To Date", "ToDate"],
    Title.BILL_DATE: ["Bill Date", "BillDate"],
    Title.BILL_NUMBER: ["Bill Number", "BillNumber"],
    Title.USAGE: ["Usage"],
    Title.PRICE: ["Price"],
    Title.AMOUNT: ["Amount"],
    Title.PRODUCT_ITEM_NAME: ["Product Item Name", "ProductItemName"],
    Title.RATE_NAME: ["Rate Name", "RateName"],
    Title.DESCRIPTION: ["Description"],
    Title.PRODUCT_ITEM_CLASS: ["Product Item Class", "ProductItemClass"],
    Title.CUSTOMER_NUMBER: ["Customer Number", "CustomerNumber"],
}


def make_column_map(title_row):
    titles = [cell.value for cell in title_row]
    column_map = {}
    for title, title_names in COLUMNS.items():
        idx = None
        for title_name in title_names:
            try:
                idx = titles.index(title_name)
            except ValueError:
                pass
        if idx is None:
            raise BadRequest(f"For the title {title} a column can't be found")

        column_map[title] = idx + 1
    return column_map


def get_date_naive(sheet, row, col):
    value = get_value(sheet, row, col)
    if value in ("", None):
        return None
    elif not isinstance(value, Datetime):
        raise BadRequest(
            f"Problem reading {value} as a timestamp at row {row} col {col}."
        )
    return value


def get_date(sheet, row, col_name):
    dt = get_date_naive(sheet, row, col_name)
    return None if dt is None else to_utc(to_ct(dt))


def get_value(sheet, row, col):
    try:
        return sheet.cell(row=row, column=col).value
    except IndexError:
        raise BadRequest(
            f"Can't find the cell at row {row} and col {col} on sheet {sheet}."
        )


def get_str(sheet, row, col):
    value = get_value(sheet, row, col)
    if value is None:
        return None

    return value.strip()


def get_dec(sheet, row, col):
    value = get_value(sheet, row, col)
    if value in ("", None):
        return None

    try:
        return Decimal(str(value))
    except InvalidOperation as e:
        raise BadRequest(
            f"Problem parsing the number '{value}' at row {row} col {col}. {e}"
        )


def get_int(sheet, row, col):
    value = get_value(sheet, row, col)
    try:
        return int(value)
    except ValueError as e:
        raise BadRequest(
            f"Problem parsing the integer '{value}' at row {row} col {col}. {e}"
        )


def _bd_add(bd, el_name, val):
    if el_name.split("-")[-1] in ("rate", "kva"):
        if el_name not in bd:
            bd[el_name] = set()
        bd[el_name].add(val)
    else:
        if el_name not in bd:
            bd[el_name] = 0
        try:
            bd[el_name] += val
        except TypeError as e:
            raise BadRequest(
                f"Problem with element name {el_name} and value '{val}': {e}"
            )


def _customer_mods(cust_number, description, bill):
    if cust_number == 10001596:
        if (
            description == "RO Mutualisation 2021-22"
            and "ro-gbp" in bill["breakdown"]
            and bill["issue_date"] == to_utc(ct_datetime(2023, 4, 14))
            and bill["start_date"] == to_utc(ct_datetime(2023, 3, 1))
            and bill["finish_date"] == to_utc(ct_datetime(2023, 3, 31, 23, 30))
        ):
            bill["start_date"] = to_utc(ct_datetime(2021, 4, 1))
            bill["finish_date"] = to_utc(ct_datetime(2022, 3, 31, 23, 30))

    return bill


def _parse_row(sheet, row, title_row):
    column_map = make_column_map(title_row)
    val = get_value(sheet, row, column_map[Title.METER_POINT])
    try:
        mpan_core = parse_mpan_core(str(int(val)))
    except ValueError as e:
        raise BadRequest(
            f"Can't parse the MPAN core in column 'Meter Point' with value '{val}' : "
            f"{e}"
        )

    bill_period = get_value(sheet, row, column_map[Title.BILL_PERIOD])
    if "-" in bill_period:
        period_start_naive, period_finish_naive = [
            Datetime.strptime(v, "%Y-%m-%d") for v in bill_period.split(" - ")
        ]
        period_start = to_utc(to_ct(period_start_naive))
        period_finish = to_utc(to_ct(period_finish_naive + relativedelta(days=1) - HH))
    else:
        period_start, period_finish = None, None

    from_date = get_date(sheet, row, column_map[Title.FROM_DATE])
    if from_date is None:
        if period_start is None:
            raise BadRequest("Can't find a bill start date.")
        else:
            from_date = period_start

    to_date_naive = get_date_naive(sheet, row, column_map[Title.TO_DATE])
    if to_date_naive is None:
        if period_finish is None:
            raise BadRequest("Can't find a bill finish date.")
        else:
            to_date = period_finish

    else:
        to_date = to_utc(to_ct(to_date_naive + relativedelta(days=1) - HH))

    issue_date = get_date(sheet, row, column_map[Title.BILL_DATE])
    bill_number = get_value(sheet, row, column_map[Title.BILL_NUMBER])
    bill = {
        "bill_type_code": "N",
        "kwh": Decimal(0),
        "vat": Decimal("0.00"),
        "net": Decimal("0.00"),
        "reads": [],
        "breakdown": {},
        "account": mpan_core,
        "issue_date": issue_date,
        "start_date": from_date,
        "finish_date": to_date,
        "mpan_core": mpan_core,
    }
    bd = bill["breakdown"]

    usage = get_dec(sheet, row, column_map[Title.USAGE])
    # usage_units = get_value(sheet, row, 'Usage Unit')
    price = get_dec(sheet, row, column_map[Title.PRICE])
    amount = get_dec(sheet, row, column_map[Title.AMOUNT])
    product_item_name = get_value(sheet, row, column_map[Title.PRODUCT_ITEM_NAME])
    rate_name = get_value(sheet, row, column_map[Title.RATE_NAME])
    if product_item_name == "Renewables Obligation (RO)" and usage is not None:
        bill["kwh"] += round(usage, 2)
    description = get_value(sheet, row, column_map[Title.DESCRIPTION])
    product_class = get_value(sheet, row, column_map[Title.PRODUCT_ITEM_CLASS])
    if description in ("Standard VAT@20%", "Reduced VAT@5%"):
        bill["vat"] += round(amount, 2)
        if description.endswith("20%"):
            vat_percentage = Decimal("20")
        else:
            vat_percentage = Decimal("5")
        bd["vat_percentage"] = vat_percentage
    else:
        bill["net"] += round(amount, 2)

        path = ["" if v is None else v for v in [product_class, description, rate_name]]
        names = _find_names(ELEM_MAP, path)

        duos_avail_prefix = "DUoS Availability ("
        duos_excess_avail_prefix = "DUoS Excess Availability ("

        if description.startswith("DUoS Availability Adjustment "):
            _bd_add(bd, "duos-availability-gbp", amount)
        elif description.startswith("DUoS Availability"):
            if description.startswith(duos_avail_prefix):
                _bd_add(
                    bd,
                    "duos-availability-kva",
                    int(description[len(duos_avail_prefix) : -5]),
                )
            _bd_add(bd, "duos-availability-days", usage)
            _bd_add(bd, "duos-availability-rate", price)
            _bd_add(bd, "duos-availability-gbp", amount)
        elif description.startswith("DUoS Excess Availability"):
            if description.startswith(duos_excess_avail_prefix):
                kva = int(description[len(duos_excess_avail_prefix) : -5])
                _bd_add(bd, "duos-excess-availability-kva", kva)
            _bd_add(bd, "duos-excess-availability-days", usage)
            _bd_add(bd, "duos-excess-availability-rate", price)
            _bd_add(bd, "duos-excess-availability-gbp", amount)
        elif description.startswith("BSUoS Black Start "):
            _bd_add(bd, "black-start-gbp", amount)
        elif description.startswith("BSUoS Reconciliation - "):
            if usage is not None:
                _bd_add(bd, "bsuos-nbp-kwh", usage)
            if price is not None:
                _bd_add(bd, "bsuos-rate", price)
            _bd_add(bd, "bsuos-gbp", amount)
        elif description.startswith("FiT Rec - "):
            _bd_add(bd, "fit-gbp", amount)
        elif description.startswith("FiT Reconciliation "):
            _bd_add(bd, "fit-gbp", amount)
        elif description.startswith("CfD FiT Rec - ") or description.startswith(
            "CfD FiT Reconciliation"
        ):
            _bd_add(bd, "cfd-fit-gbp", amount)
        elif description.startswith("Flex"):
            _bd_add(bd, "reconciliation-gbp", amount)
        elif description.startswith("Legacy TNUoS Reversal "):
            _bd_add(bd, "triad-gbp", amount)
        elif description.startswith("Hand Held Read -"):
            _bd_add(bd, "meter-rental-gbp", amount)
        elif description.startswith("RO Mutualisation "):
            _bd_add(bd, "ro-gbp", amount)
        elif description.startswith("OOC MOP - "):
            _bd_add(bd, "meter-rental-gbp", amount)
        elif description.startswith("KVa Adjustment "):
            _bd_add(bd, "duos-availability-gbp", amount)
        elif names is not None:
            for elem_k, elem_v in zip(names, (amount, price, usage)):
                if elem_k is not None:
                    _bd_add(bd, elem_k, elem_v)
        else:
            raise BadRequest(
                f"For the path {path} the parser can't work out the element."
            )

    reference = f"{bill_number}_{row}"
    for k, v in tuple(bd.items()):
        if isinstance(v, set):
            bd[k] = list(v)
        elif k.endswith("-gbp"):
            reference += "_" + k[:-4]

    bill["reference"] = reference
    bill["gross"] = bill["net"] + bill["vat"]
    customer_number = get_value(sheet, row, column_map[Title.CUSTOMER_NUMBER])
    return _customer_mods(customer_number, description, bill)


def _make_raw_bills(sheet):
    bills = []
    rows = tuple(sheet.rows)
    title_row = rows[0]
    for row_index, row in enumerate(rows[1:], start=2):
        if len(row) > 21:
            val = row[21].value
            if val not in (None, ""):
                try:
                    bills.append(_parse_row(sheet, row_index, title_row))
                except BadRequest as e:
                    raise BadRequest(f"On row {row_index}: {e.description}")

    return bills


class Parser:
    def __init__(self, f):
        self.book = load_workbook(f, data_only=True)
        self.sheet = self.book.worksheets[0]

        self.last_line = None
        lines = (self._set_last_line(i, l) for i, l in enumerate(f))
        self.reader = csv.reader(lines, skipinitialspace=True)
        self._line_number = None
        self._title_line = None

    @property
    def line_number(self):
        return None if self._line_number is None else self._line_number + 1

    def _set_last_line(self, i, line):
        self._line_numer = i
        self.last_line = line
        if i == 0:
            self._title_line = line
        return line

    def make_raw_bills(self):
        return _make_raw_bills(self.sheet)



import csv
from collections import namedtuple
from datetime import datetime as Datetime
from decimal import Decimal
from io import StringIO

from dateutil.relativedelta import relativedelta

from werkzeug.exceptions import BadRequest

from chellow.models import Session, Supply
from chellow.utils import HH, parse_mpan_core, to_ct, to_utc

READ_TYPE_MAP = {
    "A": "A",
    "C": "C",
    "D": "D",
    "E": "E",
    "F": "F",
    "X": "H",
    "I": "I",
    "R": "N",
    "O": "O",
    "Q": "Q",
    "S": "S",
    "Z": "Z",
}

SSC_MAP = {
    # 0326
    "20 0000 5855 970": {"Night": "00187", "Other": "00210", "Weekday": "00184"},
    "20 0000 6048 528": {"Night": "00187", "Other": "00210", "Weekday": "00184"},
    "20 0002 1909 377": {"Night": "00187", "Other": "00210", "Weekday": "00184"},
    "20 0002 2254 214": {"Night": "00187", "Other": "00210", "Weekday": "00184"},
    "20 0002 2326 515": {"Night": "00187", "Other": "00210", "Weekday": "00184"},
    "20 0002 2371 155": {"Night": "00187", "Other": "00210", "Weekday": "00184"},
    "20 0002 5282 171": {"Night": "00187", "Other": "00210", "Weekday": "00184"},
    "20 0002 5476 287": {"Night": "00187", "Other": "00210", "Weekday": "00184"},
    "20 0002 6228 157": {"Night": "00187", "Other": "00210", "Weekday": "00184"},
    "20 0002 6440 184": {"Night": "00187", "Other": "00210", "Weekday": "00184"},
    "20 0002 6419 768": {"Night": "00187", "Other": "00210", "Weekday": "00184"},
    # 0246
    "22 0001 3834 361": {
        "Night": "00277",
        "Other": "00160",
        "Weekday": "00276",
    },
    "22 0001 4442 321": {
        "Night": "00277",
        "Other": "00160",
        "Weekday": "00276",
    },
    "22 0002 1401 756": {
        "Night": "00277",
        "Other": "00160",
        "Weekday": "00276",
    },
    "22 0002 1427 578": {
        "Night": "00277",
        "Other": "00160",
        "Weekday": "00276",
    },
    "22 0002 1442 103": {
        "Night": "00277",
        "Other": "00160",
        "Weekday": "00276",
    },
    "22 0002 1502 222": {
        "Night": "00277",
        "Other": "00160",
        "Weekday": "00276",
    },
    "22 0002 1528 986": {
        "Night": "00277",
        "Other": "00160",
        "Weekday": "00276",
    },
    "22 0002 1540 176": {
        "Night": "00277",
        "Other": "00160",
        "Weekday": "00276",
    },
    "22 0002 1589 599": {
        "Night": "00277",
        "Other": "00160",
        "Weekday": "00276",
    },
    "22 0002 1823 142": {
        "Night": "00277",
        "Other": "00160",
        "Weekday": "00276",
    },
    "22 0002 1829 376": {
        "Night": "00277",
        "Other": "00160",
        "Weekday": "00276",
    },
    "22 0002 1839 541": {
        "Night": "00277",
        "Other": "00160",
        "Weekday": "00276",
    },
    "22 0002 1878 160": {
        "Night": "00277",
        "Other": "00160",
        "Weekday": "00276",
    },
    "22 0002 1394 727": {
        "Night": "00277",
        "Other": "00160",
        "Weekday": "00276",
    },
    # 0174
    "20 0002 6308 157": {"Day": "01072", "Night": "01071"},
    "20 0002 6467 511": {"Day": "01072", "Night": "01071"},
    "0008": {
        "Other": "00159",
    },
    "0065": {
        "Other": "00010",
    },
    "0036": {
        "Other": "00151",
    },
    "0037": {
        "Other": "00153",
    },
    "0151": {
        "Day": "00043",
        "Night": "00210",
    },
    "0154": {
        "Day": "00039",
        "Night": "00221",
    },
    "0174": {"Day": "01071", "Night": "01072"},
    "0179": {"Day": "01139", "Night": "01140"},
    "0184": {"Day": "01149", "Night": "01150"},
    "0186": {"Day": "01153", "Night": "01154"},
    "0242": {"Day": "00044", "Night": "00208"},
    "0244": {"Day": "00040", "Night": "00206"},
    "0246": {
        "Night": "00160",
        "Other": "00277",
        "Weekday": "00276",
    },
    "0252": {
        "Other": "00212",
    },
    "0265": {
        "Other": "00190",
    },
    "0319": {
        "Other": "00071",
        "Weekday": "00183",
    },
    "0320": {
        "Other": "00072",
        "Weekday": "00184",
    },
    "0322": {
        "Weekday": "01073",
        "Other": "01074",
    },
    "0326": {
        "Night": "00210",
        "Other": "00187",
        "Weekday": "00184",
    },
    "0393": {"Day": "00001"},
    "0428": {
        "Energy Charges": "00258",
        "Energy Charges 2": "00259",
    },
}


# None denotes a TPR-based charge

TITLE_MAP = {
    "Standing Charge": ("standing-gbp", "standing-rate", "standing-days"),
    "Climate Change Levy": ("ccl-gbp", "ccl-rate", "ccl-kwh"),
    "CCL": ("ccl-gbp", "ccl-rate", "ccl-kwh"),
    "Night": None,
    "Day": None,
    "Weekday": None,
    "Other": None,
    "Energy Charges": None,
}


BillElement = namedtuple("BillElement", ["gbp", "rate", "cons", "titles", "desc"])


def _to_date(row, index):
    try:
        return to_utc(to_ct(Datetime.strptime(row[index], "%Y%m%d")))
    except ValueError as e:
        raise BadRequest(f"Can't parse the date at index {index}: {e}.")


class Parser:
    def __init__(self, f):
        self.reader = csv.reader(
            StringIO(str(f.read(), "utf-8", errors="ignore")), skipinitialspace=True
        )
        self.line_number = None

    def make_raw_bills(self):
        raw_bills = []
        with Session() as sess:
            headers = {"sess": sess}

            blank_set = set(("",))
            for self.line_number, row in enumerate(self.reader):
                # skip blank lines
                if len(row) == 0 or set(row) == blank_set:
                    continue

                try:
                    bill = _process_line(row[0], row[1:], headers)
                except BadRequest as e:
                    raise BadRequest(
                        f"Can't parse line number {self.line_number}: {row} : "
                        f"{e.description}"
                    )

                if bill is not None:
                    raw_bills.append(bill)
        return raw_bills


def _process_line(code, row, headers):
    if "breakdown" in headers:
        headers["breakdown"]["raw-lines"].append(row)

    if code == "INVOICE":
        _process_INVOICE(row, headers)

    elif code == "MPAN":
        _process_MPAN(row, headers)

    elif code == "VAT":
        return _process_VAT(row, headers)

    elif code == "READING":
        _process_READING(row, headers)

    elif code == "CHARGE":
        _process_CHARGE(row, headers)

    elif code == "SUMMARY":
        _process_SUMMARY(row, headers)

    else:
        raise BadRequest(f"The code {code} is not recognized.")


def _process_INVOICE(row, headers):
    sess = headers["sess"]
    headers.clear()
    headers["kwh"] = Decimal("0")
    headers["reads"] = []
    headers["breakdown"] = {"raw-lines": []}
    headers["bill_elements"] = []
    headers["sess"] = sess

    headers["reference"] = row[0]
    bill_type_raw = row[3]
    headers["bill_type_code"] = "W" if bill_type_raw == "" else bill_type_raw
    headers["issue_date"] = _to_date(row, 14)

    headers["start_date"] = _to_date(row, 17)
    headers["finish_date"] = _to_date(row, 18) + relativedelta(days=1) - HH
    headers["net"] = Decimal("0.00") + Decimal(row[19])
    headers["vat"] = Decimal("0.00") + Decimal(row[20])
    headers["gross"] = Decimal("0.00") + Decimal(row[21])
    headers["account"] = row[26]


def _process_MPAN(row, headers):
    dno = row[3]
    unique = row[4].zfill(10)
    check_digit = row[5]

    headers["mpan_core"] = parse_mpan_core("".join([dno, unique, check_digit]))


def _process_VAT(rows, headers):
    sess = headers["sess"]
    mpan_core = headers["mpan_core"]
    start_date = headers["start_date"]
    reads = headers["reads"]
    supply = Supply.get_by_mpan_core(sess, mpan_core)
    era = supply.find_era_at(sess, start_date)
    bill_elements = []
    if era is None:
        era = supply.find_last_era(sess)
    if era is not None and era.ssc is not None:
        try:
            ssc_lookup = era.imp_mpan_core
            tpr_map = SSC_MAP[ssc_lookup]
        except KeyError:
            ssc_lookup = era.ssc.code
            try:
                tpr_map = SSC_MAP[ssc_lookup]
            except KeyError:
                raise BadRequest("The SSC " + ssc_lookup + " isn't in the SSC_MAP.")

        for read in reads:
            desc = read["tpr_code"]
            try:
                read["tpr_code"] = tpr_map[desc]
            except KeyError:
                raise BadRequest(
                    "The description " + desc + " isn't in the SSC_MAP "
                    "for the SSC " + ssc_lookup + "."
                )

        for el in headers["bill_elements"]:
            if el.titles is None:
                try:
                    tpr = tpr_map[el.desc]
                except KeyError:
                    raise BadRequest(
                        f"The billing element description {el.desc} isn't in "
                        f"the SSC_MAP for the SSC {ssc_lookup}."
                    )

                titles = (tpr + "-gbp", tpr + "-rate", tpr + "-kwh")
            else:
                titles = el.titles

            bill_elements.append(
                BillElement(
                    gbp=el.gbp, titles=titles, rate=el.rate, cons=el.cons, desc=None
                )
            )
    else:
        for read in reads:
            read["tpr_code"] = "00001"

        for el in headers["bill_elements"]:
            if el.titles is None:
                des = el.desc
                titles = (des + "-kwh", des + "-rate", des + "-gbp")
            else:
                titles = el.titles

            bill_elements.append(
                BillElement(
                    gbp=el.gbp, titles=titles, rate=el.rate, cons=el.cons, desc=None
                )
            )

    breakdown = headers["breakdown"]
    for bill_el in bill_elements:
        eln_gbp, eln_rate, eln_cons = bill_el.titles
        breakdown[eln_gbp] = bill_el.gbp
        rate = bill_el.rate
        if eln_rate is not None and rate is not None:
            try:
                rates = breakdown[eln_rate]
            except KeyError:
                rates = breakdown[eln_rate] = set()

            rates.add(rate)

        cons = bill_el.cons
        if eln_cons is not None and cons is not None:
            breakdown[eln_cons] = cons

    return {
        "kwh": headers["kwh"],
        "reference": headers["reference"],
        "mpan_core": mpan_core,
        "issue_date": headers["issue_date"],
        "account": headers["account"],
        "start_date": start_date,
        "finish_date": headers["finish_date"],
        "net": headers["net"],
        "vat": headers["vat"],
        "gross": headers["gross"],
        "breakdown": breakdown,
        "reads": reads,
        "bill_type_code": headers["bill_type_code"],
    }


def _process_READING(row, headers):
    coefficient = Decimal(row[2])
    msn = row[4]
    tpr_code_raw = row[6]

    prev_read_date = _to_date(row, 11) + relativedelta(days=1) - HH
    prev_reading_value = Decimal(row[12])
    try:
        prev_read_type = READ_TYPE_MAP[row[13]]
    except KeyError as e:
        raise BadRequest(f"The previous register read type isn't recognized {e}.")

    pres_read_date = _to_date(row, 14) + relativedelta(days=1) - HH
    pres_reading_value = Decimal(row[15])
    try:
        pres_read_type = READ_TYPE_MAP[row[16]]
    except KeyError as e:
        raise BadRequest(f"The present register read type isn't recognized {e}.")

    units_raw = row[18]
    if units_raw == "kW":
        units = "kW"
        tpr_code = None
    elif units_raw == "kVA":
        units = "kVA"
        tpr_code = None
    elif units_raw == "KWH":
        units = "kWh"
        tpr_code = tpr_code_raw
    else:
        raise BadRequest(f"Units {units_raw} not recognized.")

    headers["reads"].append(
        {
            "msn": msn,
            "mpan": "",
            "coefficient": coefficient,
            "units": units,
            "tpr_code": tpr_code,
            "prev_date": prev_read_date,
            "prev_value": prev_reading_value,
            "prev_type_code": prev_read_type,
            "pres_date": pres_read_date,
            "pres_value": pres_reading_value,
            "pres_type_code": pres_read_type,
        }
    )


def _process_CHARGE(row, headers):
    desc = row[6]

    try:
        titles = TITLE_MAP[desc]
    except KeyError:
        raise BadRequest(f"Can't find the description {desc} in the TITLE_MAP.")

    rate = Decimal(row[20])

    if desc == "Standing Charge":
        start_date = _to_date(row, 23)
        finish_date = _to_date(row, 24)
        consumption = Decimal((finish_date - start_date).days + 1)
    else:
        consumption = Decimal(row[22])

    if titles is None:
        headers["kwh"] += consumption

    gbp = Decimal("0.00") + Decimal(row[25])

    if desc == "Energy Charges":
        headers["bill_elements"].append(
            BillElement(
                gbp=(gbp / 2),
                rate=rate,
                cons=(consumption / 2),
                titles=titles,
                desc=desc,
            )
        )
        headers["bill_elements"].append(
            BillElement(
                gbp=(gbp / 2),
                rate=rate,
                cons=(consumption / 2),
                titles=titles,
                desc="Energy Charges 2",
            )
        )
    else:
        headers["bill_elements"].append(
            BillElement(gbp=gbp, rate=rate, cons=consumption, titles=titles, desc=desc)
        )


def _process_SUMMARY(row, headers):
    pass



import csv
from io import StringIO

from utils import match_tables

from chellow.models import (
    Comm,
    Contract,
    Cop,
    DtcMeterType,
    EnergisationStatus,
    GspGroup,
    MarketRole,
    MeterPaymentType,
    MeterType,
    Mtc,
    MtcLlfc,
    MtcParticipant,
    Participant,
    Pc,
    Site,
    Source,
    User,
    UserRole,
    VoltageLevel,
    insert_comms,
    insert_cops,
    insert_dtc_meter_types,
    insert_energisation_statuses,
    insert_sources,
    insert_voltage_levels,
)

from chellow.reports.report_33 import _process, content
from chellow.utils import ct_datetime, to_utc, utc_datetime


def test_content(mocker, sess):
    vf = to_utc(ct_datetime(2000, 1, 1))
    site = Site.insert(sess, "CI017", "Water Works")
    market_role_Z = MarketRole.insert(sess, "Z", "Non-core")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    participant.insert_party(sess, market_role_Z, "None core", vf, None, None)
    bank_holiday_rate_script = {"bank_holidays": []}
    Contract.insert_non_core(
        sess, "bank_holidays", "", {}, vf, None, bank_holiday_rate_script
    )
    market_role_X = MarketRole.insert(sess, "X", "Supplier")
    market_role_M = MarketRole.insert(sess, "M", "Mop")
    market_role_C = MarketRole.insert(sess, "C", "HH Dc")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    participant.insert_party(sess, market_role_M, "Fusion Mop", vf, None, None)
    participant.insert_party(sess, market_role_X, "Fusion", vf, None, None)
    participant.insert_party(sess, market_role_C, "Fusion DC", vf, None, None)
    mop_contract = Contract.insert_mop(
        sess, "Fusion", participant, "", {}, vf, None, {}
    )
    dc_contract = Contract.insert_dc(
        sess, "Fusion DC 2000", participant, "", {}, vf, None, {}
    )
    pc = Pc.insert(sess, "00", "hh", vf, None)
    insert_cops(sess)
    cop = Cop.get_by_code(sess, "5")
    insert_comms(sess)
    comm = Comm.get_by_code(sess, "GSM")
    imp_supplier_contract = Contract.insert_supplier(
        sess, "Fusion Supplier 2000", participant, "", {}, vf, None, {}
    )
    dno = participant.insert_party(sess, market_role_R, "WPD", vf, None, "22")
    meter_type = MeterType.insert(sess, "C5", "COP 1-5", vf, None)
    meter_payment_type = MeterPaymentType.insert(sess, "CR", "Credit", vf, None)
    mtc = Mtc.insert(sess, "845", False, True, vf, None)
    mtc_participant = MtcParticipant.insert(
        sess,
        mtc,
        participant,
        "HH COP5 And Above With Comms",
        False,
        True,
        meter_type,
        meter_payment_type,
        0,
        vf,
        None,
    )
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    llfc = dno.insert_llfc(
        sess, "510", "PC 5-8 & HH HV", voltage_level, False, True, vf, None
    )
    MtcLlfc.insert(sess, mtc_participant, llfc, vf, None)
    insert_sources(sess)
    source = Source.get_by_code(sess, "grid")
    insert_energisation_statuses(sess)
    energisation_status = EnergisationStatus.get_by_code(sess, "E")
    gsp_group = GspGroup.insert(sess, "_L", "South Western")
    insert_dtc_meter_types(sess)
    dtc_meter_type = DtcMeterType.get_by_code(sess, "H")
    supply = site.insert_e_supply(
        sess,
        source,
        None,
        "Bob",
        utc_datetime(2000, 1, 1),
        None,
        gsp_group,
        mop_contract,
        dc_contract,
        "hgjeyhuw",
        dno,
        pc,
        "845",
        cop,
        comm,
        None,
        energisation_status,
        dtc_meter_type,
        "22 7867 6232 781",
        "510",
        imp_supplier_contract,
        "7748",
        361,
        None,
        None,
        None,
        None,
        None,
    )
    editor = UserRole.insert(sess, "editor")
    user = User.insert(sess, "admin@example.com", "xxx", editor, None)
    user_id = user.id
    supply_id = supply.id

    sess.commit()
    f = StringIO()
    mocker.patch("chellow.reports.report_33.open_file", return_value=f)
    date = utc_datetime(2024, 1, 1)
    mpan_cores = None
    content(user_id, date, supply_id, mpan_cores)


def test_process_2010(mocker, sess):
    vf = to_utc(ct_datetime(2000, 1, 1))
    site = Site.insert(sess, "CI017", "Water Works")
    market_role_Z = MarketRole.insert(sess, "Z", "Non-core")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    participant.insert_party(sess, market_role_Z, "None core", vf, None, None)
    bank_holiday_rate_script = {"bank_holidays": []}
    Contract.insert_non_core(
        sess, "bank_holidays", "", {}, vf, None, bank_holiday_rate_script
    )
    market_role_X = MarketRole.insert(sess, "X", "Supplier")
    market_role_M = MarketRole.insert(sess, "M", "Mop")
    market_role_C = MarketRole.insert(sess, "C", "HH Dc")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    participant.insert_party(sess, market_role_M, "Fusion Mop", vf, None, None)
    participant.insert_party(sess, market_role_X, "Fusion", vf, None, None)
    participant.insert_party(sess, market_role_C, "Fusion DC", vf, None, None)
    mop_contract = Contract.insert_mop(
        sess, "Fusion", participant, "", {}, vf, None, {}
    )
    dc_contract = Contract.insert_dc(
        sess, "Fusion DC 2000", participant, "", {}, vf, None, {}
    )
    pc = Pc.insert(sess, "00", "hh", vf, None)
    insert_cops(sess)
    cop = Cop.get_by_code(sess, "5")
    insert_comms(sess)
    comm = Comm.get_by_code(sess, "GSM")
    imp_supplier_contract = Contract.insert_supplier(
        sess, "Fusion Supplier 2000", participant, "", {}, vf, None, {}
    )
    dno = participant.insert_party(sess, market_role_R, "WPD", vf, None, "22")
    Contract.insert_dno(
        sess,
        dno.dno_code,
        participant,
        "",
        {},
        vf,
        None,
        {
            "tariffs": {"510": {"day-gbp-per-kwh": 1, "night-gbp-per-kwh": 1}},
            "lafs": {"hv": {"other": 1}},
        },
    )
    meter_type = MeterType.insert(sess, "C5", "COP 1-5", utc_datetime(2000, 1, 1), None)
    meter_payment_type = MeterPaymentType.insert(sess, "CR", "Credit", vf, None)
    mtc = Mtc.insert(sess, "845", False, True, vf, None)
    mtc_participant = MtcParticipant.insert(
        sess,
        mtc,
        participant,
        "HH COP5 And Above With Comms",
        False,
        True,
        meter_type,
        meter_payment_type,
        0,
        vf,
        None,
    )
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    llfc = dno.insert_llfc(
        sess,
        "510",
        "PC 5-8 & HH HV",
        voltage_level,
        False,
        True,
        utc_datetime(1996, 1, 1),
        None,
    )
    MtcLlfc.insert(sess, mtc_participant, llfc, vf, None)
    insert_sources(sess)
    source = Source.get_by_code(sess, "grid")
    insert_energisation_statuses(sess)
    energisation_status = EnergisationStatus.get_by_code(sess, "E")
    gsp_group = GspGroup.insert(sess, "_L", "South Western")
    insert_dtc_meter_types(sess)
    dtc_meter_type = DtcMeterType.get_by_code(sess, "H")
    supply = site.insert_e_supply(
        sess,
        source,
        None,
        "Bob",
        utc_datetime(2000, 1, 1),
        None,
        gsp_group,
        mop_contract,
        dc_contract,
        "hgjeyhuw",
        dno,
        pc,
        "845",
        cop,
        comm,
        None,
        energisation_status,
        dtc_meter_type,
        "22 7867 6232 781",
        "510",
        imp_supplier_contract,
        "7748",
        361,
        None,
        None,
        None,
        None,
        None,
    )

    sess.commit()
    f = StringIO()
    mocker.patch("chellow.reports.report_33.open_file", return_value=f)
    date = utc_datetime(2010, 1, 1)
    supply_id = supply.id
    mpan_cores = None
    _process(sess, f, date, supply_id, mpan_cores)
    f.seek(0)
    actual = list(csv.reader(f))
    expected = [
        [
            "Date",
            "Import MPAN Core",
            "Export MPAN Core",
            "Physical Site Id",
            "Physical Site Name",
            "Other Site Ids",
            "Other Site Names",
            "Supply Id",
            "Source",
            "Generator Type",
            "GSP Group",
            "DNO Name",
            "Voltage Level",
            "Is Substations",
            "Metering Type",
            "Mandatory HH",
            "PC",
            "MTC",
            "CoP",
            "Comms Type",
            "SSC Code",
            "SSC Description",
            "Energisation Status",
            "Number Of Registers",
            "MOP Contract",
            "DC Contract",
            "Meter Serial Number",
            "Meter Installation Date",
            "Latest Normal Meter Read Date",
            "Latest Normal Meter Read Type",
            "Latest DC Bill Date",
            "Latest MOP Bill Date",
            "Supply Start Date",
            "Supply Finish Date",
            "DTC Meter Type",
            "tnuos_band",
            "Import ACTIVE?",
            "Import REACTIVE_IMPORT?",
            "Import REACTIVE_EXPORT?",
            "Export ACTIVE?",
            "Export REACTIVE_IMPORT?",
            "Export REACTIVE_EXPORT?",
            "Import Agreed Supply Capacity (kVA)",
            "Import LLFC Code",
            "Import LLFC Description",
            "Import Supplier Contract",
            "Import Supplier Account",
            "Import Mandatory kW",
            "Latest Import Supplier Bill Date",
            "Export Agreed Supply Capacity (kVA)",
            "Export LLFC Code",
            "Export LLFC Description",
            "Export Supplier Contract",
            "Export Supplier Account",
            "Export Mandatory kW",
            "Latest Export Supplier Bill Date",
        ],
        [
            "2010-01-01 00:00",
            "22 7867 6232 781",
            "",
            "CI017",
            "Water Works",
            "",
            "",
            "1",
            "grid",
            "",
            "_L",
            "22",
            "HV",
            "False",
            "hh",
            "no",
            "00",
            "845",
            "5",
            "GSM",
            "",
            "",
            "E",
            "",
            "Fusion",
            "Fusion DC 2000",
            "hgjeyhuw",
            "2000-01-01 00:00",
            "hh",
            "",
            "",
            "",
            "2000-01-01 00:00",
            "",
            "H",
            "",
            "false",
            "false",
            "false",
            "false",
            "false",
            "false",
            "361",
            "510",
            "PC 5-8 & HH HV",
            "Fusion Supplier 2000",
            "7748",
            "",
            "",
            "",
            "",
            "",
            "",
            "",
            "",
            "",
        ],
    ]
    match_tables(expected, actual)


def test_process_2024(mocker, sess):
    vf = to_utc(ct_datetime(2000, 1, 1))
    site = Site.insert(sess, "CI017", "Water Works")
    market_role_Z = MarketRole.insert(sess, "Z", "Non-core")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    participant.insert_party(sess, market_role_Z, "None core", vf, None, None)
    bank_holiday_rate_script = {"bank_holidays": []}
    Contract.insert_non_core(
        sess, "bank_holidays", "", {}, vf, None, bank_holiday_rate_script
    )
    market_role_X = MarketRole.insert(sess, "X", "Supplier")
    market_role_M = MarketRole.insert(sess, "M", "Mop")
    market_role_C = MarketRole.insert(sess, "C", "HH Dc")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    participant.insert_party(sess, market_role_M, "Fusion Mop", vf, None, None)
    participant.insert_party(sess, market_role_X, "Fusion", vf, None, None)
    participant.insert_party(sess, market_role_C, "Fusion DC", vf, None, None)
    mop_contract = Contract.insert_mop(
        sess, "Fusion", participant, "", {}, vf, None, {}
    )
    dc_contract = Contract.insert_dc(
        sess, "Fusion DC 2000", participant, "", {}, vf, None, {}
    )
    pc = Pc.insert(sess, "00", "hh", vf, None)
    insert_cops(sess)
    cop = Cop.get_by_code(sess, "5")
    insert_comms(sess)
    comm = Comm.get_by_code(sess, "GSM")
    imp_supplier_contract = Contract.insert_supplier(
        sess, "Fusion Supplier 2000", participant, "", {}, vf, None, {}
    )
    dno = participant.insert_party(sess, market_role_R, "WPD", vf, None, "22")
    Contract.insert_dno(
        sess,
        dno.dno_code,
        participant,
        "",
        {},
        vf,
        None,
        {
            "_L": {
                "bands": {},
                "tariffs": {
                    "510": {
                        "description": "HV Site Specific Band 1",
                        "gbp-per-kvarh": 1,
                        "green-gbp-per-kwh": 1,
                    }
                },
            },
        },
    )
    meter_type = MeterType.insert(sess, "C5", "COP 1-5", utc_datetime(2000, 1, 1), None)
    meter_payment_type = MeterPaymentType.insert(sess, "CR", "Credit", vf, None)
    mtc = Mtc.insert(sess, "845", False, True, vf, None)
    mtc_participant = MtcParticipant.insert(
        sess,
        mtc,
        participant,
        "HH COP5 And Above With Comms",
        False,
        True,
        meter_type,
        meter_payment_type,
        0,
        vf,
        None,
    )
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    llfc = dno.insert_llfc(
        sess,
        "510",
        "PC 5-8 & HH HV",
        voltage_level,
        False,
        True,
        utc_datetime(1996, 1, 1),
        None,
    )
    MtcLlfc.insert(sess, mtc_participant, llfc, vf, None)
    insert_sources(sess)
    source = Source.get_by_code(sess, "grid")
    insert_energisation_statuses(sess)
    energisation_status = EnergisationStatus.get_by_code(sess, "E")
    gsp_group = GspGroup.insert(sess, "_L", "South Western")
    insert_dtc_meter_types(sess)
    dtc_meter_type = DtcMeterType.get_by_code(sess, "H")
    supply = site.insert_e_supply(
        sess,
        source,
        None,
        "Bob",
        utc_datetime(2000, 1, 1),
        None,
        gsp_group,
        mop_contract,
        dc_contract,
        "hgjeyhuw",
        dno,
        pc,
        "845",
        cop,
        comm,
        None,
        energisation_status,
        dtc_meter_type,
        "22 7867 6232 781",
        "510",
        imp_supplier_contract,
        "7748",
        361,
        None,
        None,
        None,
        None,
        None,
    )

    sess.commit()
    f = StringIO()
    mocker.patch("chellow.reports.report_33.open_file", return_value=f)
    date = utc_datetime(2024, 1, 1)
    supply_id = supply.id
    mpan_cores = None
    _process(sess, f, date, supply_id, mpan_cores)
    f.seek(0)
    actual = list(csv.reader(f))
    expected = [
        [
            "Date",
            "Import MPAN Core",
            "Export MPAN Core",
            "Physical Site Id",
            "Physical Site Name",
            "Other Site Ids",
            "Other Site Names",
            "Supply Id",
            "Source",
            "Generator Type",
            "GSP Group",
            "DNO Name",
            "Voltage Level",
            "Is Substations",
            "Metering Type",
            "Mandatory HH",
            "PC",
            "MTC",
            "CoP",
            "Comms Type",
            "SSC Code",
            "SSC Description",
            "Energisation Status",
            "Number Of Registers",
            "MOP Contract",
            "DC Contract",
            "Meter Serial Number",
            "Meter Installation Date",
            "Latest Normal Meter Read Date",
            "Latest Normal Meter Read Type",
            "Latest DC Bill Date",
            "Latest MOP Bill Date",
            "Supply Start Date",
            "Supply Finish Date",
            "DTC Meter Type",
            "tnuos_band",
            "Import ACTIVE?",
            "Import REACTIVE_IMPORT?",
            "Import REACTIVE_EXPORT?",
            "Export ACTIVE?",
            "Export REACTIVE_IMPORT?",
            "Export REACTIVE_EXPORT?",
            "Import Agreed Supply Capacity (kVA)",
            "Import LLFC Code",
            "Import LLFC Description",
            "Import Supplier Contract",
            "Import Supplier Account",
            "Import Mandatory kW",
            "Latest Import Supplier Bill Date",
            "Export Agreed Supply Capacity (kVA)",
            "Export LLFC Code",
            "Export LLFC Description",
            "Export Supplier Contract",
            "Export Supplier Account",
            "Export Mandatory kW",
            "Latest Export Supplier Bill Date",
        ],
        [
            "2024-01-01 00:00",
            "22 7867 6232 781",
            "",
            "CI017",
            "Water Works",
            "",
            "",
            "1",
            "grid",
            "",
            "_L",
            "22",
            "HV",
            "False",
            "hh",
            "no",
            "00",
            "845",
            "5",
            "GSM",
            "",
            "",
            "E",
            "",
            "Fusion",
            "Fusion DC 2000",
            "hgjeyhuw",
            "2000-01-01 00:00",
            "hh",
            "",
            "",
            "",
            "2000-01-01 00:00",
            "",
            "H",
            "HV1",
            "false",
            "false",
            "false",
            "false",
            "false",
            "false",
            "361",
            "510",
            "PC 5-8 & HH HV",
            "Fusion Supplier 2000",
            "7748",
            "",
            "",
            "",
            "",
            "",
            "",
            "",
            "",
            "",
        ],
    ]
    match_tables(expected, actual)



import atexit
import collections
import csv
import threading
import traceback
from datetime import timedelta
from importlib import import_module
from io import StringIO

import requests

from werkzeug.exceptions import BadRequest

from chellow.models import (
    Contract,
    Session,
)
from chellow.utils import (
    ct_datetime_now,
    ct_datetime_parse,
    date_format,
    to_utc,
    utc_datetime_now,
)

importer = None


def api_get(path, params=None):
    url = f"https://api.neso.energy/api/3/action/{path}"
    res = requests.get(url, params=params, timeout=120)
    try:
        res_j = res.json()
    except requests.exceptions.JSONDecodeError as e:
        raise BadRequest(
            f"Couldn't parse as JSON the content from {url} with error {e}: "
            f"{res.text}"
        )

    if "success" in res_j and not res_j["success"]:
        raise BadRequest(res_j)

    return res_j


def csv_download(url):
    res = requests.get(url, timeout=120)
    csv_file = StringIO(res.text)
    return csv.DictReader(csv_file)


def csv_get(resource_id):
    res_j = api_get("resource_show", params={"id": resource_id})
    download_url = res_j["result"]["url"]
    return csv_download(download_url)


def csv_latest(package_id, last_import_date, name=None):
    params = {"id": package_id}
    res_j = api_get("datapackage_show", params=params)
    latest_lm = None
    latest_resource = None
    for resource in res_j["result"]["resources"]:
        if name is None or resource["name"].startswith(name):
            last_modified = resource["last_modified"]
            if last_import_date is None or last_modified > last_import_date:
                if latest_lm is None or last_modified > latest_lm:
                    latest_lm = last_modified
                    latest_resource = resource

    if latest_resource is None:
        return None
    else:
        return csv_download(latest_resource["path"])


def parse_date(date_str):
    if "/" in date_str:
        fmt = "%d/%m/%Y"
    elif "-" in date_str:
        fmt = "%Y-%m-%d"
    else:
        raise BadRequest(f"Couldn't parse the date {date_str}")

    return to_utc(ct_datetime_parse(date_str, fmt))


def run_import(sess, log, set_progress):
    log("Starting to import data from the NESO")

    for mod_name in (
        "chellow.e.aahedc",
        "chellow.e.tnuos",
        "chellow.e.bsuos",
        "chellow.e.triad",
    ):
        mod = import_module(mod_name)
        mod.neso_import(sess, log, set_progress)


LAST_RUN_KEY = "last_run"
GLOBAL_ALERT = "There's a problem with a <a href='/e/neso'>NESO import</a>."
NG_STATE_KEY = "neso"


class Neso(threading.Thread):
    def __init__(self):
        super().__init__(name="NESO")
        self.messages = collections.deque(maxlen=500)
        self.progress = ""
        self.stopped = threading.Event()
        self.going = threading.Event()
        self.global_alert = None

    def stop(self):
        self.stopped.set()
        self.going.set()
        self.join()

    def go(self):
        self.going.set()

    def log(self, message):
        self.messages.appendleft(
            f"{ct_datetime_now().strftime('%Y-%m-%d %H:%M:%S')} - {message}"
        )

    def set_progress(self, progress):
        self.progress = progress

    def run(self):
        while not self.stopped.is_set():
            with Session() as sess:
                try:
                    config = Contract.get_non_core_by_name(sess, "configuration")
                    state = config.make_state()
                    ng_state = state.get(NG_STATE_KEY, {})
                except BaseException as e:
                    msg = f"{e.description} " if isinstance(e, BadRequest) else ""
                    self.log(f"{msg}{traceback.format_exc()}")
                    self.global_alert = GLOBAL_ALERT
                    sess.rollback()

            last_run = ng_state.get(LAST_RUN_KEY)
            if last_run is None or utc_datetime_now() - last_run > timedelta(days=1):
                self.going.set()

            if self.going.is_set():
                self.global_alert = None
                with Session() as sess:
                    try:
                        config = Contract.get_non_core_by_name(sess, "configuration")
                        state = config.make_state()
                        try:
                            ng_state = state[NG_STATE_KEY]
                        except KeyError:
                            ng_state = state[NG_STATE_KEY] = {}

                        ng_state[LAST_RUN_KEY] = utc_datetime_now()
                        config.update_state(state)
                        sess.commit()
                        run_import(sess, self.log, self.set_progress)
                    except BaseException as e:
                        msg = f"{e.description} " if isinstance(e, BadRequest) else ""
                        self.log(f"{msg}{traceback.format_exc()}")
                        self.global_alert = GLOBAL_ALERT
                        sess.rollback()
                    finally:
                        self.going.clear()
                        self.log("Finished importing NESO data.")

            else:
                self.log(
                    f"The importer was last run at {date_format(last_run)}. There will "
                    f"be another import when 24 hours have elapsed since the last run."
                )
                self.going.wait(60 * 60)


def get_importer():
    return importer


def startup():
    global importer
    importer = Neso()
    importer.start()


@atexit.register
def shutdown():
    if importer is not None:
        importer.stop()



import csv
import itertools
from codecs import iterdecode
from datetime import datetime as Datetime
from decimal import Decimal, InvalidOperation

from werkzeug.exceptions import BadRequest

from chellow.utils import parse_mpan_core, to_utc, validate_hh_start

# "Time stamp","Value","Events","Comment","User"
# "18/10/2024 09:30:00","88.0","","",""
# "18/10/2024 09:00:00","89.1","","",""


def create_parser(reader, mpan_map, messages):
    return HhParserSchneiderCsv(reader, mpan_map, messages)


def get_field(values, index, name):
    if len(values) > index:
        return values[index].strip()
    else:
        raise BadRequest(f"Can't find field {index}, {name}.")


def parse_values(values):
    start_date_str = get_field(values, 0, "Timestamp")
    start_date = validate_hh_start(
        to_utc(Datetime.strptime(start_date_str, "%d/%m/%Y %H:%M:%S"))
    )

    reading_str = get_field(values, 1, "Reading")
    reading_str = reading_str.replace(",", "")
    try:
        reading = Decimal(reading_str)
    except InvalidOperation as e:
        raise BadRequest(f"Problem parsing the number {reading_str}. {e}")
    return start_date, reading


class HhParserSchneiderCsv:
    def __init__(self, reader, mpan_map, messages):
        s = iterdecode(reader, "utf-8")
        self.shredder = zip(itertools.count(1), csv.reader(s))
        next(self.shredder)  # skip the title line
        self.line_number, self.values = next(self.shredder)
        _, self.pres_reading = parse_values(self.values)
        self.mpan_core = parse_mpan_core(next(iter(mpan_map.values())))

    def __iter__(self):
        return self

    def __next__(self):
        try:
            self.line_number, self.values = next(self.shredder)
            start_date, reading = parse_values(self.values)
            datum = {
                "mpan_core": self.mpan_core,
                "channel_type": "ACTIVE",
                "start_date": start_date,
                "value": self.pres_reading - reading,
                "status": "A",
            }
            self.pres_reading = reading
            return datum
        except BadRequest as e:
            e.description = (
                f"Problem at line number: {self.line_number}: {self.values}: "
                f"{e.description}"
            )
            raise e

    def close(self):
        self.shredder.close()



import atexit
import collections
import threading
import traceback
from datetime import timedelta

from sqlalchemy import false, null, select

from werkzeug.exceptions import BadRequest

from chellow.e.computer import contract_func
from chellow.gas.engine import g_contract_func
from chellow.models import (
    Batch,
    BillType,
    Contract,
    GBatch,
    GBill,
    GContract,
    GRateScript,
    GReadType,
    GSupply,
    GUnit,
    RateScript,
    ReadType,
    Session,
    Supply,
    Tpr,
)
from chellow.utils import (
    c_months_u,
    ct_datetime_now,
    date_format,
    keydefaultdict,
    utc_datetime_now,
)

importer = None


def run_import(sess, log, set_progress):
    log("Starting to update the fake batches")
    caches = {}

    now_ct = ct_datetime_now()
    (last_month_start, last_month_finish), (
        current_month_start,
        current_month_finish,
    ) = list(c_months_u(finish_year=now_ct.year, finish_month=now_ct.month, months=2))

    for last_rate_script in sess.scalars(
        select(RateScript)
        .join(Contract, RateScript.contract_id == Contract.id)
        .where(RateScript.finish_date == null())
    ):
        contract = last_rate_script.contract
        fb_func = contract_func(caches, contract, "make_fake_bills")
        if fb_func is None:
            continue

        fake_batch_name = f"fake_e_batch_{contract.id}"

        fake_batch = sess.scalar_one_or_none(
            select(Batch).where(
                Batch.contract == contract,
                Batch.reference == fake_batch_name,
            )
        )

        if fake_batch is not None and fake_batch.start_date < current_month_start:
            fake_batch.delete(sess)
            sess.flush()
            fake_batch = None

        if fake_batch is None:
            fake_batch = contract.insert_batch(sess, fake_batch_name, "Fake Batch")
            bill_types = keydefaultdict(lambda k: BillType.get_by_code(sess, k))

            tprs = keydefaultdict(
                lambda k: None if k is None else Tpr.get_by_code(sess, k)
            )

            read_types = keydefaultdict(lambda k: ReadType.get_by_code(sess, k))
            for raw_bill in fb_func(
                sess,
                log,
                last_month_start,
                last_month_finish,
                current_month_start,
                current_month_finish,
            ):
                mpan_core = raw_bill["mpan_core"]
                supply = Supply.get_by_mpan_core(sess, mpan_core)
                bill = fake_batch.insert_bill(
                    sess,
                    raw_bill["account"],
                    raw_bill["reference"],
                    raw_bill["issue_date"],
                    raw_bill["start_date"],
                    raw_bill["finish_date"],
                    raw_bill["kwh"],
                    raw_bill["net"],
                    raw_bill["vat"],
                    raw_bill["gross"],
                    bill_types[raw_bill["bill_type_code"]],
                    raw_bill["breakdown"],
                    supply,
                )
                for raw_read in raw_bill["reads"]:
                    bill.insert_read(
                        sess,
                        tprs[raw_read["tpr_code"]],
                        raw_read["coefficient"],
                        raw_read["units"],
                        raw_read["msn"],
                        raw_read["mpan"],
                        raw_read["prev_date"],
                        raw_read["prev_value"],
                        read_types[raw_read["prev_type_code"]],
                        raw_read["pres_date"],
                        raw_read["pres_value"],
                        read_types[raw_read["pres_type_code"]],
                    )
    for last_rate_script in sess.scalars(
        select(GRateScript)
        .join(GContract, GRateScript.g_contract_id == GContract.id)
        .where(GContract.is_industry == false(), GRateScript.finish_date == null())
    ):
        g_contract = last_rate_script.g_contract
        log(f"Looking at gas contract {g_contract.name}")
        fb_func = g_contract_func(caches, g_contract, "make_fake_bills")
        if fb_func is None:
            log("Doesn't have a make_fake_bills function so skipping")
            continue

        fake_batch_name = f"fake_g_batch_{g_contract.id}"

        fake_batch = sess.scalars(
            select(GBatch).where(
                GBatch.g_contract == g_contract,
                GBatch.reference == fake_batch_name,
            )
        ).one_or_none()

        if fake_batch is not None:
            log(f"Found existing fake batch {fake_batch_name}")
            first_fake_bill = sess.scalars(
                select(GBill)
                .where(GBill.g_batch == fake_batch)
                .order_by(GBill.start_date)
            ).first()
            if (
                first_fake_bill is None
                or first_fake_bill.start_date < current_month_start
            ):
                fake_batch.delete(sess)
                sess.flush()
                fake_batch = None
                log(f"Deleted fake batch {fake_batch_name}")

        if fake_batch is None:
            latest_bill = sess.scalars(
                select(GBill)
                .join(GBatch)
                .where(GBatch.g_contract == g_contract)
                .order_by(GBill.finish_date.desc())
            ).first()
            if latest_bill is None:
                log(
                    "No actual bills found for this contract, so not creating fake "
                    "batch"
                )
                continue
            if latest_bill.finish_date >= last_month_start:
                log("Actual bills found in the last month, so not creating fake batch")
                continue

            raw_bills = fb_func(
                sess,
                log,
                last_month_start,
                last_month_finish,
                current_month_start,
                current_month_finish,
            )
            if raw_bills is not None and len(raw_bills) > 0:
                log(f"Adding a new fake batch {fake_batch_name}")
                fake_batch = g_contract.insert_g_batch(
                    sess, fake_batch_name, "Fake Batch"
                )
                log("About to insert raw bills")
                for raw_bill in raw_bills:
                    bill_type = BillType.get_by_code(sess, raw_bill["bill_type_code"])
                    g_supply = GSupply.get_by_mprn(sess, raw_bill["mprn"])
                    g_bill = fake_batch.insert_g_bill(
                        sess,
                        g_supply,
                        bill_type,
                        raw_bill["reference"],
                        raw_bill["account"],
                        raw_bill["issue_date"],
                        raw_bill["start_date"],
                        raw_bill["finish_date"],
                        raw_bill["kwh"],
                        raw_bill["net_gbp"],
                        raw_bill["vat_gbp"],
                        raw_bill["gross_gbp"],
                        raw_bill["raw_lines"],
                        raw_bill["breakdown"],
                    )
                    sess.flush()
                    for raw_read in raw_bill["reads"]:
                        prev_type = GReadType.get_by_code(
                            sess, raw_read["prev_type_code"]
                        )
                        pres_type = GReadType.get_by_code(
                            sess, raw_read["pres_type_code"]
                        )
                        g_unit = GUnit.get_by_code(sess, raw_read["unit"])
                        g_bill.insert_g_read(
                            sess,
                            raw_read["msn"],
                            g_unit,
                            raw_read["correction_factor"],
                            raw_read["calorific_value"],
                            raw_read["prev_value"],
                            raw_read["prev_date"],
                            prev_type,
                            raw_read["pres_value"],
                            raw_read["pres_date"],
                            pres_type,
                        )
                    sess.commit()
            else:
                log("No raw bills available, so not adding fake batch")
            sess.commit()


LAST_RUN_KEY = "fake_batch_updater_last_run"


class FakeBatchUpdater(threading.Thread):
    def __init__(self):
        super().__init__(name="Fake Batch Updater")
        self.messages = collections.deque(maxlen=500)
        self.progress = ""
        self.stopped = threading.Event()
        self.going = threading.Event()
        self.global_alert = None

    def stop(self):
        self.stopped.set()
        self.going.set()
        self.join()

    def go(self):
        self.going.set()

    def log(self, message):
        self.messages.appendleft(
            f"{ct_datetime_now().strftime('%Y-%m-%d %H:%M:%S')} - {message}"
        )

    def set_progress(self, progress):
        self.progress = progress

    def run(self):
        while not self.stopped.is_set():
            with Session() as sess:
                try:
                    config = Contract.get_non_core_by_name(sess, "configuration")
                    state = config.make_state()
                except BaseException as e:
                    msg = f"{e.description} " if isinstance(e, BadRequest) else ""
                    self.log(f"{msg}{traceback.format_exc()}")
                    self.global_alert = (
                        "There's a problem with a <a href='/fake_batch_updater'>"
                        "Fake Batch Updater</a>."
                    )
                    sess.rollback()

            last_run = state.get(LAST_RUN_KEY)
            if last_run is None or utc_datetime_now() - last_run > timedelta(days=1):
                self.going.set()

            if self.going.is_set():
                self.global_alert = None
                with Session() as sess:
                    try:
                        config = Contract.get_non_core_by_name(sess, "configuration")
                        state = config.make_state()
                        state[LAST_RUN_KEY] = utc_datetime_now()
                        config.update_state(state)
                        sess.commit()
                        run_import(sess, self.log, self.set_progress)
                    except BaseException as e:
                        msg = f"{e.description} " if isinstance(e, BadRequest) else ""
                        self.log(f"{msg}{traceback.format_exc()}")
                        self.global_alert = (
                            "There's a problem with a "
                            "<a href='/fake_batch_updater'>Fake Batch Updater</a>."
                        )
                        sess.rollback()
                    finally:
                        self.going.clear()
                        self.log("Finished updating fake batches.")

            else:
                self.log(
                    f"The updater was last run at {date_format(last_run)}. There will "
                    f"be another update when a hour has elapsed since the last run."
                )
                self.going.wait(60 * 60)


def get_importer():
    return importer


def startup():
    global importer
    importer = FakeBatchUpdater()
    importer.start()


@atexit.register
def shutdown():
    if importer is not None:
        importer.stop()
